"""Base model."""

from ansys.api.edb.v1.edb_messages_pb2 import EDBObjMessage


class ObjBase:
    """Provides the base object that all gRPC-related models extend from."""

    def __init__(self, msg):
        """Initialize the base object.

        Parameters
        ----------
        msg : EDBObjMessage
        """
        self._id = 0 if msg is None else msg.id

    @property
    def is_null(self):
        """:obj:`bool`: Flag indicating if the object exists in the database.

        This property is read-only.
        """
        return self.id == 0

    @property
    def id(self):
        """:obj:`int`: Unique ID of the EDB object.

        A ``0`` indicates an invalid object.

        This property is read-only.
        """
        return self._id

    @property
    def msg(self):
        """:obj:`EDBObjMessage`: Protobuf message that represents the object's ID.

        This property can only be set to ``None``.
        """
        return EDBObjMessage(id=self.id)

    @msg.setter
    def msg(self, val):
        if val is None:
            self._id = 0


class TypeField(object):
    """Provides a descriptor for a type field that can be overridden by subclasses.

    You can have an optional ``@property def _type(self)`` in the same class as a fallback method to
    fetch the type from the server in case the static type is unknown.

    Examples
    --------
    class PrimitiveType(Enum):
        CIRCLE = 0

    class Primitive(object):
        type = TypeField(None)
        @property
        def _type(self):
            return PrimitiveType(self.__stub.GetPrimitiveType())

    class Circle(Primitive):
        type = TypeField(PrimitiveType.CIRCLE)

    Primitive(...).type # => fetch type from server.
    Primitive.type      # => None.
    Circle(...).type    # => CIRCLE.
    Circle.type         # => CIRCLE.
    """

    def __init__(self, tp):
        """Initialize a type field with static concrete type."""
        self._type = tp

    def __get__(self, instance, owner):
        """Get the static type if possible. Otherwise, fetch and return the instance type."""
        if self._type is not None:
            # return static type
            return self._type
        if instance is not None and hasattr(instance, "_type"):
            # fetch type
            return instance._type  # noqa
        else:
            # no type info available
            return None



"""This module contains utility functions for API development work."""


def map_list(iterable_to_operate_on, operator=None):
    """Apply the given operator to each member of an iterable and return the modified list.

    Parameters
    ---------
    iterable_to_operate on
    operator
    """
    return list(
        iterable_to_operate_on if operator is None else map(operator, iterable_to_operate_on)
    )



"""Layout."""

from ansys.api.edb.v1 import layout_pb2
from ansys.api.edb.v1.layout_pb2_grpc import LayoutServiceStub

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.hierarchy import cell_instance, group, pin_group
from ansys.edb.core.inner import ObjBase, messages, parser, utils, variable_server
from ansys.edb.core.layer.layer_collection import LayerCollection
from ansys.edb.core.layout import voltage_regulator
from ansys.edb.core.layout.mcad_model import McadModel
from ansys.edb.core.layout_instance import layout_instance
from ansys.edb.core.net.differential_pair import DifferentialPair
from ansys.edb.core.net.extended_net import ExtendedNet
from ansys.edb.core.net.net import Net
from ansys.edb.core.net.net_class import NetClass
from ansys.edb.core.primitive.primitive import BoardBendDef, PadstackInstance, Primitive
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.terminal.terminals import Terminal


class Layout(ObjBase, variable_server.VariableServer):
    """Represents a layout."""

    __stub: LayoutServiceStub = StubAccessor(StubType.layout)

    def __init__(self, msg):
        """Initialize a new layout.

        Parameters
        ----------
        msg : EDBObjMessage
        """
        ObjBase.__init__(self, msg)
        variable_server.VariableServer.__init__(self, msg)

    @property
    def cell(self):
        """:class:`.Cell`: Owning cell for the layout.

        This property is read-only.
        """
        from ansys.edb.core.layout.cell import Cell

        return Cell(self.__stub.GetCell(self.msg))

    @property
    def layer_collection(self):
        """:class:`.LayerCollection`: \
        Layer collection of the layout."""
        return LayerCollection(self.__stub.GetLayerCollection(self.msg))

    @layer_collection.setter
    def layer_collection(self, layer_collection):
        self.__stub.SetLayerCollection(
            layout_pb2.SetLayerCollectionMessage(
                layout=self.msg, layer_collection=layer_collection.msg
            )
        )

    def _get_items(self, obj_type, lyt_obj_type_enum, do_cast=False):
        """Get a list of layout objects."""
        items = utils.map_list(
            self.__stub.GetItems(messages.layout_get_items_message(self, lyt_obj_type_enum)).items,
            obj_type,
        )
        return items if not do_cast else [item.cast() for item in items]

    @property
    def primitives(self):
        """:obj:`list` of :class:`.Primitive`: List of all \
            primitives in the layout.

        This property is read-only.
        """
        return self._get_items(Primitive, LayoutObjType.PRIMITIVE, True)

    @property
    def padstack_instances(self):
        """:obj:`list` of :class:`.PadstackInstance`: List of \
            all padstack instances in the layout.

        This property is read-only.
        """
        return self._get_items(PadstackInstance, LayoutObjType.PADSTACK_INSTANCE)

    @property
    def terminals(self):
        """:obj:`list` of :class:`.Terminal`: \
        List of all terminals in the layout.

        This property is read-only.
        """
        return self._get_items(Terminal, LayoutObjType.TERMINAL, True)

    @property
    def cell_instances(self):
        """:obj:`list` of :class:`CellInstances <.CellInstance>`: \
        List of all cell instances in the layout.

        This property is read-only.
        """
        return self._get_items(cell_instance.CellInstance, LayoutObjType.CELL_INSTANCE)

    @property
    def nets(self):
        """:obj:`list` of :class:`.Net`: List of all nets \
        in the layout.

        This property is read-only.
        """
        return self._get_items(Net, LayoutObjType.NET)

    @property
    def groups(self):
        """:obj:`list` of :class:`.Group`: List of all groups \
        in the layout.

        This property is read-only.
        """
        return self._get_items(group.Group, LayoutObjType.GROUP, True)

    @property
    def net_classes(self):
        """:obj:`list` of :class:`.NetClass`: List of all \
        net classes in the layout.

        This property is read-only.
        """
        return self._get_items(NetClass, LayoutObjType.NET_CLASS)

    @property
    def differential_pairs(self):
        """:obj:`list` of :class:`.DifferentialPair`: \
        List of all differential pairs in the layout.

        This property is read-only.
        """
        return self._get_items(DifferentialPair, LayoutObjType.DIFFERENTIAL_PAIR)

    @property
    def pin_groups(self):
        """:obj:`list` of :class:`.PinGroup` : List of all \
        pin groups in the layout.

        This property is read-only.
        """
        return self._get_items(pin_group.PinGroup, LayoutObjType.PIN_GROUP)

    @property
    def voltage_regulators(self):
        """:obj:`list` of :class:`.VoltageRegulator`: \
        List of all voltage regulators in the layout.

        This property is read-only.
        """
        return self._get_items(voltage_regulator.VoltageRegulator, LayoutObjType.VOLTAGE_REGULATOR)

    @property
    def extended_nets(self):
        """:obj:`list` of :class:`.ExtendedNet`: \
        List of all extended nets in the layout.

        This property is read-only.
        """
        return self._get_items(ExtendedNet, LayoutObjType.EXTENDED_NET)

    @parser.to_polygon_data
    def expanded_extent(
        self, nets, extent, expansion_factor, expansion_unitless, use_round_corner, num_increments
    ):
        """Get an expanded polygon for a list of nets.

        Parameters
        ----------
        nets : list[:class:`.Net`]
            List of nets.
        extent : :class:`.ExtentType`
            Geometry extent type for expansion.
        expansion_factor : float
            Expansion factor for the polygon union. No expansion occurs if the value
            for this parameter is less than or equal to 0.
        expansion_unitless : bool
            When unitless, the distance by which the extent expands is the factor
            multiplied by the longer dimension (X or Y distance) of the expanded
            object/net.
        use_round_corner : bool
            Whether to use round corners or sharp corners. For round corners, this
            returns a bounding box if its area is within 10% of the rounded expansion's
            area.
        num_increments : int
            Number of iterations desired to reach the full expansion.

        Returns
        -------
        :class:`.PolygonData`

        Notes
        -----
        This method returns the expansion of the contour, so any voids within expanded
        objects are ignored.
        """
        return self.__stub.GetExpandedExtentFromNets(
            messages.layout_expanded_extent_message(
                self,
                nets,
                extent,
                expansion_factor,
                expansion_unitless,
                use_round_corner,
                num_increments,
            )
        )

    def convert_primitives_to_vias(self, primitives, is_pins=False):
        """Convert a list of primitives into vias or pins.

        Parameters
        ----------
        primitives : list[:class:`.Primitive`]
            List of primitives.
        is_pins : bool, default: False
            Whether the list consists of pins. The default is ``False``, in which
            case the list consists of vias.
        """
        self.__stub.ConvertPrimitivesToVias(
            messages.layout_convert_p2v_message(self, primitives, is_pins)
        )

    @property
    def port_reference_terminals_connected(self):
        """:obj:`bool`: Flag indicating if port reference terminals are connected.

        This property applies to lumped ports and circuit ports. It is ``True`` if
        port terminals are connected, ``False`` otherwise.

        This property is read-only.
        """
        return self.__stub.ArePortReferenceTerminalsConnected(self.msg).is_connected

    @property
    def zone_primitives(self):
        """:obj:`list` of :class:`.Primitive`: List of \
        all primitives in the :term:`zones <Zone>`.

        This property is read-only.
        """
        return [Primitive(msg) for msg in self.__stub.GetZonePrimitives(self.msg)]

    @property
    def fixed_zone_primitive(self):
        """:class:`.Primitive`: Fixed :term:`zones <Zone>` primitive."""
        msg = self.__stub.GetFixedZonePrimitive(self.msg)
        return None if msg is None else Primitive(msg).cast()

    @fixed_zone_primitive.setter
    def fixed_zone_primitive(self, value):
        self.__stub.SetFixedZonePrimitives(messages.pointer_property_message(self, value))

    @property
    def board_bend_defs(self):
        """:obj:`list` of :class:`.BoardBendDef`: List of all \
        board bend definitions in the layout.

        This property is read-only.
        """
        return [BoardBendDef(msg) for msg in self.__stub.GetBoardBendDefs(self.msg)]

    def synchronize_bend_manager(self):
        """Synchronize the bend manager."""
        self.__stub.SynchronizeBendManager(self.msg)

    @property
    def layout_instance(self):
        """:class:`.LayoutInstance`: \
        Instance of the layout.

        This property is read-only.
        """
        return layout_instance.LayoutInstance(self.__stub.GetLayoutInstance(self.msg))

    def create_stride(self, filename):
        """Create a Stride model from an MCAD file.

        Parameters
        ----------
        filename : str
            Absolute path of the MCAD file.

        Returns
        -------
        McadModel
           Stride model created.
        """
        return McadModel.create_stride(layout=self, filename=filename)

    def create_hfss(self, filename, design):
        """Create an HFSS model from an MCAD file.

        Parameters
        ----------
        filename : str
            Absolute path of the MCAD file.
        design : str
            Design name.

        Returns
        -------
        McadModel
            HFSS model created.
        """
        return McadModel.create_hfss(connectable=self, filename=filename, design=design)

    def create_3d_comp(self, filename):
        """Create a 3D composite model from an MCAD file.

        Parameters
        ----------
        filename : str
            Absolute path of the MCAD file.

        Returns
        -------
        McadModel
           3D composite model created.
        """
        return McadModel.create_3d_comp(layout=self, filename=filename)



"""IO component property."""

from ansys.api.edb.v1.io_component_property_pb2_grpc import IOComponentPropertyServiceStub
import google.protobuf.empty_pb2 as empty_pb2

from ansys.edb.core.definition import component_property, port_property, solder_ball_property
from ansys.edb.core.inner import messages
from ansys.edb.core.session import StubAccessor, StubType


class IOComponentProperty(component_property.ComponentProperty):
    """Represents an I0 component property."""

    __stub: IOComponentPropertyServiceStub = StubAccessor(StubType.io_component_property)

    @classmethod
    def create(cls):
        """
        Create an IO component property.

        Returns
        -------
        IOComponentProperty
            IO component property created.
        """
        return IOComponentProperty(cls.__stub.Create(empty_pb2.Empty()))

    @property
    def solder_ball_property(self):
        """:obj:`SolderBallProperty`: Solder ball property.

        A copy is returned. Use the setter for any modifications to be reflected.
        """
        return solder_ball_property.SolderBallProperty(
            self.__stub.GetSolderBallProperty(messages.edb_obj_message(self))
        )

    @solder_ball_property.setter
    def solder_ball_property(self, value):
        self.__stub.SetSolderBallProperty(
            messages.pointer_property_message(target=self, value=value)
        )

    @property
    def port_property(self):
        """:obj:`PortProperty`: Port property.

        A copy is returned. Use the setter for any modifications to be reflected.
        """
        return port_property.PortProperty(
            self.__stub.GetPortProperty(messages.edb_obj_message(self))
        )

    @port_property.setter
    def port_property(self, value):
        self.__stub.SetPortProperty(messages.pointer_property_message(target=self, value=value))



"""3D transformformations."""
import ansys.api.edb.v1.transform3d_pb2 as pb
from ansys.api.edb.v1.transform3d_pb2_grpc import Transform3DServiceStub
from google.protobuf import empty_pb2

from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.inner.parser import to_3_point3d_data, to_point3d_data
from ansys.edb.core.session import StubAccessor, StubType


class Transform3D(ObjBase):
    """Represents a 3D transformation.

    Parameters
    ----------
    anchor : :term:`Point3DLike`
    rot_axis_from : :term:`Point3DLike`
    rot_axis_to : :term:`Point3DLike`
    rot_angle : str, int, float, complex, Value
        Rotation angle, specified counter-clockwise in radians, from the ``rot_axis_from`` parameter
        towards the ``rot_axis_to`` parameter.
    offset : :term:`Point3DLike`
    mirror : bool
        Mirror against the YZ plane.
    """

    __stub: Transform3DServiceStub = StubAccessor(StubType.transform3d)

    @classmethod
    def create_identity(cls):
        """Create an identity transformation 3D matrix.

        Returns
        -------
        Transform3D
        """
        return Transform3D(cls.__stub.CreateIdentity(empty_pb2.Empty()))

    @classmethod
    def create_copy(cls, transform3d):
        """Create a 3D transformation by copying another 3D transformation.

        Returns
        -------
        Transform3D
        """
        return Transform3D(cls.__stub.CreateCopy(messages.edb_obj_message(transform3d)))

    @classmethod
    def create_from_matrix(cls, matrix):
        """Create a 3D transformation from general matrix data.

        Parameters
        ----------
        matrix : :obj:`list` of :obj:`list` of :obj:`floats`
            Array (4x4) to copy from.

        Returns
        -------
        Transform3D
        """
        trans = Transform3D(cls.__stub.CreateIdentity(empty_pb2.Empty()))
        trans.matrix = matrix
        return trans

    @classmethod
    def create_from_offset(cls, offset):
        """Create a 3D transformation with an offset.

        Parameters
        ----------
        offset : :term:`Point3DLike`
            Vector offset.
        Returns
        -------
        Transform3D
        """
        return Transform3D(cls.__stub.CreateOffset(messages.cpos_3d_message(offset)))

    @classmethod
    def create_from_center_scale(cls, center, scale):
        """Create a 3D transformation for scaling about a point.

        Parameters
        ----------
        center : :term:`Point3DLike`
            Center of the transformation.
        scale : :obj:`float`
            Scale factor of the transformation.

        Returns
        -------
        Transform3D
        """
        return Transform3D(
            cls.__stub.CreateCenterScale(messages.cpos_3d_double_message(center, scale))
        )

    @classmethod
    def create_from_angle(cls, zyx_decomposition):
        """Create a 3D transformation from ZYX decomposition.

        Parameters
        ----------
        zyx_decomposition : :term:`Point3DLike`
            ZYX decomposition.

        Returns
        -------
        Transform3D
        """
        return Transform3D(
            cls.__stub.CreateRotationFromAngle(messages.cpos_3d_message(zyx_decomposition))
        )

    @classmethod
    def create_from_axis(cls, x, y, z):
        """Create a 3D transformation with a rotation matrix from three axes.

        Parameters
        ----------
        x : :term:`Point3DLike`
            X axis.
        y : :term:`Point3DLike`
            Y axis.
        z : :term:`Point3DLike`
            Z axis.

        Returns
        -------
        Transform3D
        """
        return Transform3D(
            cls.__stub.CreateRotationFromAxis(messages.cpos_3d_triple_message(x, y, z))
        )

    @classmethod
    def create_from_axis_and_angle(cls, axis, angle):
        """Create a 3D transformation with the given axis and angle.

        Parameters
        ----------
        axis : :term:`Point3DLike`
            Axis.
        angle : :obj:`float`
            Angle.

        Returns
        -------
        Transform3D
        """
        return Transform3D(
            cls.__stub.CreateRotationFromAxisAndAngle(messages.cpos_3d_double_message(axis, angle))
        )

    @classmethod
    def create_from_one_axis_to_another(cls, from_axis, to_axis):
        """Create a 3D transformformation with rotation from an axis to an axis.

        Parameters
        ----------
        from_axis : :term:`Point3DLike`
            From axis.
        to_axis : :term:`Point3DLike`
            To axis.

        Returns
        -------
        Transform3D
        """
        return Transform3D(
            cls.__stub.CreateRotationFromToAxis(messages.cpos_3d_pair_message(from_axis, to_axis))
        )

    @classmethod
    def create_from_transform_2d(cls, transform, z_off):
        """Create a 3D transformation with transform data.

        Parameters
        ----------
        transform : :class:`.Transform`
            Transform data.
        z_off : :obj:`float`
            Z offset.

        Returns
        -------
        Transform3D
        """
        return Transform3D(
            cls.__stub.CreateTransform2D(messages.double_property_message(transform, z_off))
        )

    def transpose(self):
        """Transpose the 3D transformation."""
        self.__stub.Transpose(messages.edb_obj_message(self))

    def invert(self):
        """Invert the 3D transformation."""
        self.__stub.Invert(messages.edb_obj_message(self))

    def is_identity(self, eps, rotation):
        """Get identity of the 3D transformation.

        Parameters
        ----------
        rotation : :obj:`bool`
        eps : :obj:`float`

        Returns
        -------
        :obj:`bool`
        """
        return self.__stub.IsIdentity(
            pb.IsIdentityMessage(
                target=messages.edb_obj_message(self),
                eps=eps,
                rotation=rotation,
            )
        ).value

    def is_equal(self, other_transform, rotation, eps):
        """Equality check for two #D transformations.

        Parameters
        ----------
        other_transform
        rotation : :obj:`bool`
        eps : :obj:`float`

        Returns
        -------
        :obj:`bool`
            Result of equality check.

        """
        return self.__stub.IsEqual(
            pb.IsEqualMessage(
                target=messages.edb_obj_message(self),
                value=messages.edb_obj_message(other_transform),
                eps=eps,
                rotation=rotation,
            )
        ).value

    def __add__(self, other_transform):
        """Add operator and concatenate two 3D transformations.

        Parameters
        ----------
        other_transform : Transform3D
            Second 3D transformation.

        Returns
        -------
        Transform
            3D transformation object created.
        """
        return Transform3D(
            self.__stub.OperatorPlus(messages.pointer_property_message(self, other_transform))
        )

    @property
    @to_3_point3d_data
    def axis(self):
        """:obj:`list` of :term:`Point3DLike`: Axis."""
        return self.__stub.GetAxis(messages.edb_obj_message(self))

    @to_point3d_data
    def transform_point(self, point):
        """Get the transform point of the 3D transformation.

        Parameters
        ----------
        point : :term:`Point3DLike`

        Returns
        -------
        :term:`Point3DLike`
            Transform point.
        """
        return self.__stub.TransformPoint(messages.cpos_3d_property_message(self, point))

    @property
    @to_point3d_data
    def z_y_x_rotation(self):
        """:term:`Point3DLike`: ZYX rotation."""
        return self.__stub.GetZYXRotation(messages.edb_obj_message(self))

    @property
    @to_point3d_data
    def scaling(self):
        """:term:`Point3DLike`: Scaling."""
        return self.__stub.GetScaling(messages.edb_obj_message(self))

    @property
    @to_point3d_data
    def shift(self):
        """:term:`Point3DLike`: Shift."""
        return self.__stub.GetShift(messages.edb_obj_message(self))

    @property
    def matrix(self):
        """:obj:`list` of :obj:`list` of :obj:`floats` : Transformation matrix as a 2D 4x4 array."""
        msg = self.__stub.GetMatrix(messages.edb_obj_message(self))
        matrix = [[float(_) for _ in msg.doubles[(i - 1) * 4 : i * 4]] for i in range(1, 5)]
        return matrix

    @matrix.setter
    def matrix(self, value):
        if (
            len(value) == 4
            and len(value[0]) == 4
            and len(value[1]) == 4
            and len(value[2]) == 4
            and len(value[3]) == 4
        ):
            unrolled_matrix = [float(j) for submatrix in value for j in submatrix]
            self.__stub.SetMatrix(messages.doubles_property_message(self, unrolled_matrix))



"""Session manager for gRPC."""
from contextlib import contextmanager
from enum import Enum
from shutil import which
from struct import pack, unpack
import subprocess
from sys import modules

from ansys.api.edb.v1.arc_data_pb2_grpc import ArcDataServiceStub
from ansys.api.edb.v1.board_bend_def_pb2_grpc import BoardBendDefServiceStub
from ansys.api.edb.v1.bondwire_def_pb2_grpc import (
    ApdBondwireDefServiceStub,
    BondwireDefServiceStub,
    Jedec4BondwireDefServiceStub,
    Jedec5BondwireDefServiceStub,
)
from ansys.api.edb.v1.bondwire_pb2_grpc import BondwireServiceStub
from ansys.api.edb.v1.bundle_term_pb2_grpc import BundleTerminalServiceStub
from ansys.api.edb.v1.cell_instance_pb2_grpc import CellInstanceServiceStub
from ansys.api.edb.v1.cell_pb2_grpc import CellServiceStub
from ansys.api.edb.v1.circle_pb2_grpc import CircleServiceStub
from ansys.api.edb.v1.component_def_pb2_grpc import ComponentDefServiceStub
from ansys.api.edb.v1.component_group_pb2_grpc import ComponentGroupServiceStub
from ansys.api.edb.v1.component_model_pb2_grpc import (
    ComponentModelServiceStub,
    DynamicLinkComponentModelServiceStub,
    NPortComponentModelServiceStub,
)
from ansys.api.edb.v1.component_pin_pb2_grpc import ComponentPinServiceStub
from ansys.api.edb.v1.component_property_pb2_grpc import ComponentPropertyServiceStub
from ansys.api.edb.v1.connectable_pb2_grpc import ConnectableServiceStub
from ansys.api.edb.v1.database_pb2_grpc import DatabaseServiceStub
from ansys.api.edb.v1.dataset_def_pb2_grpc import DatasetDefServiceStub
from ansys.api.edb.v1.debye_model_pb2_grpc import DebyeModelServiceStub
from ansys.api.edb.v1.die_property_pb2_grpc import DiePropertyServiceStub
from ansys.api.edb.v1.dielectric_material_model_pb2_grpc import DielectricMaterialModelServiceStub
from ansys.api.edb.v1.differential_pair_pb2_grpc import DifferentialPairServiceStub
from ansys.api.edb.v1.djordjecvic_sarkar_model_pb2_grpc import DjordjecvicSarkarModelServiceStub
from ansys.api.edb.v1.edge_term_pb2_grpc import EdgeServiceStub, EdgeTerminalServiceStub
from ansys.api.edb.v1.extended_net_pb2_grpc import ExtendedNetServiceStub
from ansys.api.edb.v1.group_pb2_grpc import GroupServiceStub
from ansys.api.edb.v1.hfss_simulation_settings_pb2_grpc import (
    DCRSettingsServiceStub,
    HFSSAdvancedMeshingSettingsServiceStub,
    HFSSAdvancedSettingsServiceStub,
    HFSSGeneralSettingsServiceStub,
    HFSSOptionsSettingsServiceStub,
    HFSSSolverSettingsServiceStub,
)
from ansys.api.edb.v1.hfss_simulation_setup_pb2_grpc import HfssSimulationSetupServiceStub
from ansys.api.edb.v1.hierarchy_obj_pb2_grpc import HierarchyObjectServiceStub
from ansys.api.edb.v1.ic_component_property_pb2_grpc import ICComponentPropertyServiceStub
from ansys.api.edb.v1.inst_array_pb2_grpc import InstArrayServiceStub
from ansys.api.edb.v1.io_component_property_pb2_grpc import IOComponentPropertyServiceStub
from ansys.api.edb.v1.layer_collection_pb2_grpc import LayerCollectionServiceStub
from ansys.api.edb.v1.layer_map_pb2_grpc import LayerMapServiceStub
from ansys.api.edb.v1.layer_pb2_grpc import LayerServiceStub
from ansys.api.edb.v1.layout_component_pb2_grpc import LayoutComponentServiceStub
from ansys.api.edb.v1.layout_instance_context_pb2_grpc import LayoutInstanceContextServiceStub
from ansys.api.edb.v1.layout_instance_pb2_grpc import LayoutInstanceServiceStub
from ansys.api.edb.v1.layout_obj_instance_2d_geometry_pb2_grpc import (
    LayoutObjInstance2DGeometryServiceStub,
)
from ansys.api.edb.v1.layout_obj_instance_3d_geometry_pb2_grpc import (
    LayoutObjInstance3DGeometryServiceStub,
)
from ansys.api.edb.v1.layout_obj_instance_geometry_pb2_grpc import (
    LayoutObjInstanceGeometryServiceStub,
)
from ansys.api.edb.v1.layout_obj_instance_pb2_grpc import LayoutObjInstanceServiceStub
from ansys.api.edb.v1.layout_obj_pb2_grpc import LayoutObjServiceStub
from ansys.api.edb.v1.layout_pb2_grpc import LayoutServiceStub
from ansys.api.edb.v1.material_def_pb2_grpc import MaterialDefServiceStub
from ansys.api.edb.v1.material_property_thermal_modifier_pb2_grpc import (
    MaterialPropertyThermalModifierServiceStub,
)
from ansys.api.edb.v1.mcad_model_pb2_grpc import McadModelServiceStub
from ansys.api.edb.v1.model_pb2_grpc import ModelServiceStub
from ansys.api.edb.v1.multipole_debye_model_pb2_grpc import MultipoleDebyeModelServiceStub
from ansys.api.edb.v1.net_pb2_grpc import NetServiceStub
from ansys.api.edb.v1.netclass_pb2_grpc import NetClassServiceStub
from ansys.api.edb.v1.netlist_model_pb2_grpc import NetlistModelServiceStub
from ansys.api.edb.v1.package_def_pb2_grpc import PackageDefServiceStub
from ansys.api.edb.v1.padstack_def_data_pb2_grpc import PadstackDefDataServiceStub
from ansys.api.edb.v1.padstack_def_pb2_grpc import PadstackDefServiceStub
from ansys.api.edb.v1.padstack_inst_term_pb2_grpc import PadstackInstanceTerminalServiceStub
from ansys.api.edb.v1.padstack_instance_pb2_grpc import PadstackInstanceServiceStub
from ansys.api.edb.v1.path_pb2_grpc import PathServiceStub
from ansys.api.edb.v1.pin_group_pb2_grpc import PinGroupServiceStub
from ansys.api.edb.v1.pin_group_term_pb2_grpc import PinGroupTerminalServiceStub
from ansys.api.edb.v1.pin_pair_model_pb2_grpc import PinPairModelServiceStub
from ansys.api.edb.v1.point_data_pb2_grpc import PointDataServiceStub
from ansys.api.edb.v1.point_term_pb2_grpc import PointTerminalServiceStub
from ansys.api.edb.v1.polygon_data_pb2_grpc import PolygonDataServiceStub
from ansys.api.edb.v1.polygon_pb2_grpc import PolygonServiceStub
from ansys.api.edb.v1.port_property_pb2_grpc import PortPropertyServiceStub
from ansys.api.edb.v1.primitive_pb2_grpc import PrimitiveServiceStub
from ansys.api.edb.v1.r_tree_pb2_grpc import RTreeServiceStub
from ansys.api.edb.v1.raptor_x_simulation_settings_pb2_grpc import (
    RaptorXAdvancedSettingsServiceStub,
    RaptorXGeneralSettingsServiceStub,
)
from ansys.api.edb.v1.rectangle_pb2_grpc import RectangleServiceStub
from ansys.api.edb.v1.rlc_component_property_pb2_grpc import RLCComponentPropertyServiceStub
from ansys.api.edb.v1.simulation_settings_pb2_grpc import (
    AdvancedMeshingSettingsServiceStub,
    AdvancedSettingsServiceStub,
    SettingsOptionsServiceStub,
    SimulationSettingsServiceStub,
    SolverSettingsServiceStub,
)
from ansys.api.edb.v1.simulation_setup_pb2_grpc import SimulationSetupServiceStub
from ansys.api.edb.v1.siwave_dcir_simulation_settings_pb2_grpc import (
    SIWaveDCIRSimulationSettingsServiceStub,
)
from ansys.api.edb.v1.siwave_simulation_settings_pb2_grpc import (
    SIWaveAdvancedSettingsServiceStub,
    SIWaveDCAdvancedSettingsServiceStub,
    SIWaveDCSettingsServiceStub,
    SIWaveGeneralSettingsServiceStub,
    SIWaveSParameterSettingsServiceStub,
)
from ansys.api.edb.v1.solder_ball_property_pb2_grpc import SolderBallPropertyServiceStub
from ansys.api.edb.v1.sparameter_model_pb2_grpc import SParameterModelServiceStub
from ansys.api.edb.v1.spice_model_pb2_grpc import SpiceModelServiceStub
from ansys.api.edb.v1.stackup_layer_pb2_grpc import StackupLayerServiceStub
from ansys.api.edb.v1.structure3d_pb2_grpc import Structure3DServiceStub
from ansys.api.edb.v1.term_inst_pb2_grpc import TerminalInstanceServiceStub
from ansys.api.edb.v1.term_inst_term_pb2_grpc import TerminalInstanceTerminalServiceStub
from ansys.api.edb.v1.term_pb2_grpc import TerminalServiceStub
from ansys.api.edb.v1.text_pb2_grpc import TextServiceStub
from ansys.api.edb.v1.transform3d_pb2_grpc import Transform3DServiceStub
from ansys.api.edb.v1.transform_pb2_grpc import TransformServiceStub
from ansys.api.edb.v1.value_pb2_grpc import ValueServiceStub
from ansys.api.edb.v1.variable_server_pb2_grpc import VariableServerServiceStub
from ansys.api.edb.v1.via_group_pb2_grpc import ViaGroupServiceStub
from ansys.api.edb.v1.via_layer_pb2_grpc import ViaLayerServiceStub
from ansys.api.edb.v1.voltage_regulator_pb2_grpc import VoltageRegulatorServiceStub
import grpc

from ansys.edb.core.inner import LOGGER
from ansys.edb.core.inner.exceptions import EDBSessionException, ErrorCode
from ansys.edb.core.inner.interceptors import ExceptionInterceptor, LoggingInterceptor

# The session module singleton
MOD = modules[__name__]
MOD.current_session = None


class StubAccessor(object):
    """Provides a descriptor for assignig a specific stub to a model."""

    def __init__(self, stub_type):
        """Initialize a descriptor stub with a name and stub service.

        Parameters
        ----------
        stub_type : StubType
        """
        self.__stub_name = stub_type.name

    def __get__(self, instance=None, owner=None):
        """Get the corresponding stub service if a session is active."""
        if MOD.current_session is not None:
            return MOD.current_session.stub(self.__stub_name)
        raise EDBSessionException(ErrorCode.NO_SESSIONS)


# Helper class for storing data used by the session
class _Session:
    def __init__(self, ip_address, port_num, ansys_em_root):
        if MOD.current_session is not None:
            raise EDBSessionException(ErrorCode.STARTUP_MULTI_SESSIONS)

        self.ip_address = ip_address or "localhost"
        self.port_num = port_num
        self.ansys_em_root = ansys_em_root
        self.channel = None
        self.local_server_proc = None
        self.stubs = None
        self.session = None
        self.interceptors = [
            # on reversed order of interception
            ExceptionInterceptor(LOGGER),
            LoggingInterceptor(LOGGER),
        ]

    def __enter__(self):
        if MOD.current_session == self:
            self.connect()
            return self
        else:
            raise EDBSessionException(ErrorCode.STARTUP_MULTI_SESSIONS)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def _initialize_stubs(self):
        self.stubs = {stub.name: stub.value(self.channel) for stub in StubType}

    @property
    def server_url(self):
        return "{}:{}".format(self.ip_address, self.port_num)

    @property
    def server_executable(self):
        if self.is_launch():
            return which(cmd="EDB_RPC_Server", path=self.ansys_em_root)
        else:
            return None

    def server_arguments(self):
        args = []

        if self.port_num is not None:
            args.append("-p")
            args.append(str(self.port_num))

        return args

    def stub(self, name):
        if self.is_active():
            return self.stubs.get(name)

    def is_active(self):
        return self.channel is not None and self.stubs is not None

    def is_local(self):
        return self.ip_address == "localhost"

    def is_launch(self):
        return self.ansys_em_root is not None

    def connect(self):
        if self.is_active():
            return

        if self.is_launch():
            self.start_server()

        self.channel = grpc.insecure_channel(self.server_url)
        self.channel = grpc.intercept_channel(self.channel, *self.interceptors)

        self._initialize_stubs()

    def disconnect(self):
        if self.stubs is not None:
            self.stubs = None

        if self.channel is not None:
            self.channel.close()
            self.channel = None

        if MOD.current_session == self:
            MOD.current_session = None

        if self.is_launch():
            self.stop_server()

    def start_server(self):
        if not self.is_local():
            return None

        if not self.is_launch():
            return None

        if self.server_executable is None:
            raise EDBSessionException(ErrorCode.STARTUP_NO_EXECUTABLE)

        cmd = [self.server_executable] + self.server_arguments()

        try:
            self.local_server_proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        except Exception as e:
            self.disconnect()
            raise EDBSessionException(ErrorCode.STARTUP_UNEXPECTED, e)

        self.wait_server_ready()

    def wait_server_ready(self):
        """Wait for the server to say it successfully started up."""
        local_server_proc_output = self.local_server_proc.stdout.readline()
        stdout = local_server_proc_output.decode().rstrip()

        if not stdout.startswith("Server listening on 127.0.0.1:"):
            try:
                print("Local server failed to start properly. Trying to shut down gracefully...")
                if self.local_server_proc.wait(10):
                    self._on_server_startup_error()
            except subprocess.TimeoutExpired:
                raise EDBSessionException(ErrorCode.STARTUP_TIMEOUT, stdout)
            finally:
                self.disconnect()

    def _on_server_startup_error(self):
        # Check if the server process has exited
        if not self.local_server_proc.poll():
            return

        # Get the return code in a usable format
        ret_code = self.local_server_proc.returncode
        ret_code = ret_code if ret_code < 0 else unpack("i", pack("I", ret_code))[0]

        code = (
            ErrorCode.STARTUP_FAILURE_EDB
            if ret_code == -1
            else ErrorCode.STARTUP_FAILURE_LICENSE
            if ret_code == -2
            else ErrorCode.STARTUP_FAILURE
        )
        raise EDBSessionException(code)

    def stop_server(self):
        if self.local_server_proc and not self.local_server_proc.poll():
            self.local_server_proc.terminate()
            self.local_server_proc.wait()
        self.local_server_proc = None


class StubType(Enum):
    """Provides an enum representing available service stub types."""

    cell = CellServiceStub
    database = DatabaseServiceStub
    layer_collection = LayerCollectionServiceStub
    layer = LayerServiceStub
    stackup_layer = StackupLayerServiceStub
    via_layer = ViaLayerServiceStub
    layout = LayoutServiceStub
    material = MaterialDefServiceStub
    net = NetServiceStub
    primitive = PrimitiveServiceStub
    polygon = PolygonServiceStub
    polygon_data = PolygonDataServiceStub
    path = PathServiceStub
    rectangle = RectangleServiceStub
    via_group = ViaGroupServiceStub
    circle = CircleServiceStub
    text = TextServiceStub
    terminal = TerminalServiceStub
    terminal_instance = TerminalInstanceServiceStub
    terminal_instance_terminal = TerminalInstanceTerminalServiceStub
    bundle_terminal = BundleTerminalServiceStub
    edge = EdgeServiceStub
    edge_terminal = EdgeTerminalServiceStub
    point_terminal = PointTerminalServiceStub
    padstack_instance_terminal = PadstackInstanceTerminalServiceStub
    pin_group = PinGroupServiceStub
    pin_group_terminal = PinGroupTerminalServiceStub
    bondwire = BondwireServiceStub
    bondwire_def = BondwireDefServiceStub
    apd_bondwire_def = ApdBondwireDefServiceStub
    jedec4_bondwire_def = Jedec4BondwireDefServiceStub
    jedec5_bondwire_def = Jedec5BondwireDefServiceStub
    padstack_def = PadstackDefServiceStub
    value = ValueServiceStub
    variable_server = VariableServerServiceStub
    cell_instance = CellInstanceServiceStub
    inst_array = InstArrayServiceStub
    hierarchy_obj = HierarchyObjectServiceStub
    group = GroupServiceStub
    netclass = NetClassServiceStub
    layer_map = LayerMapServiceStub
    point_data = PointDataServiceStub
    arc_data = ArcDataServiceStub
    padstack_instance = PadstackInstanceServiceStub
    voltage_regulator = VoltageRegulatorServiceStub
    connectable = ConnectableServiceStub
    component_group = ComponentGroupServiceStub
    layout_obj = LayoutObjServiceStub
    structure3d = Structure3DServiceStub
    layout_instance = LayoutInstanceServiceStub
    layout_instance_context = LayoutInstanceContextServiceStub
    layout_obj_instance = LayoutObjInstanceServiceStub
    layout_obj_instance_geometry = LayoutObjInstanceGeometryServiceStub
    layout_obj_instance_2d_geometry = LayoutObjInstance2DGeometryServiceStub
    layout_obj_instance_3d_geometry = LayoutObjInstance3DGeometryServiceStub
    component_def = ComponentDefServiceStub
    component_pin = ComponentPinServiceStub
    component_model = ComponentModelServiceStub
    nport_component_model = NPortComponentModelServiceStub
    dyn_link_component_model = DynamicLinkComponentModelServiceStub
    extended_net = ExtendedNetServiceStub
    padstack_def_data = PadstackDefDataServiceStub
    differential_pair = DifferentialPairServiceStub
    solder_ball_property = SolderBallPropertyServiceStub
    component_property = ComponentPropertyServiceStub
    ic_component_property = ICComponentPropertyServiceStub
    die_property = DiePropertyServiceStub
    port_property = PortPropertyServiceStub
    dataset_def = DatasetDefServiceStub
    package_def = PackageDefServiceStub
    dielectric_material_model = DielectricMaterialModelServiceStub
    debye_model = DebyeModelServiceStub
    multipole_debye_model = MultipoleDebyeModelServiceStub
    djordecvic_sarkar_model = DjordjecvicSarkarModelServiceStub
    mcad_model = McadModelServiceStub
    material_property_thermal_modifier = MaterialPropertyThermalModifierServiceStub
    r_tree = RTreeServiceStub
    pin_pair_model = PinPairModelServiceStub
    board_bend_def = BoardBendDefServiceStub
    sparameter_model = SParameterModelServiceStub
    spice_model = SpiceModelServiceStub
    netlist_model = NetlistModelServiceStub
    model = ModelServiceStub
    transform = TransformServiceStub
    io_component_property = IOComponentPropertyServiceStub
    rlc_component_property = RLCComponentPropertyServiceStub
    transform3d = Transform3DServiceStub
    hfss_general_sim_settings = HFSSGeneralSettingsServiceStub
    hfss_options_sim_settings = HFSSOptionsSettingsServiceStub
    hfss_advanced_sim_settings = HFSSAdvancedSettingsServiceStub
    hfss_advanced_sim_meshing_settings = HFSSAdvancedMeshingSettingsServiceStub
    hfss_solver_sim_settings = HFSSSolverSettingsServiceStub
    hfss_dcr_sim_settings = DCRSettingsServiceStub
    hfss_sim_setup = HfssSimulationSetupServiceStub
    sim_setup = SimulationSetupServiceStub
    sim_settings = SimulationSettingsServiceStub
    sim_settings_options = SettingsOptionsServiceStub
    advanced_sim_settings = AdvancedSettingsServiceStub
    advanced_mesh_sim_settings = AdvancedMeshingSettingsServiceStub
    solver_sim_settings = SolverSettingsServiceStub
    siwave_general_sim_settings = SIWaveGeneralSettingsServiceStub
    siwave_advanced_sim_settings = SIWaveAdvancedSettingsServiceStub
    siwave_dc_sim_settings = SIWaveDCSettingsServiceStub
    siwave_dc_advanced_sim_settings = SIWaveDCAdvancedSettingsServiceStub
    siwave_s_param_sim_settings = SIWaveSParameterSettingsServiceStub
    siwave_dcir_sim_settings = SIWaveDCIRSimulationSettingsServiceStub
    raptor_x_general_sim_settings = RaptorXGeneralSettingsServiceStub
    raptor_x_adv_sim_settings = RaptorXAdvancedSettingsServiceStub
    layout_component = LayoutComponentServiceStub


def attach_session(ip_address=None, port_num=50051):
    """Attach a session to a port running the EDB API server.

    Parameters
    ----------
    ip_address : str, default: None
        IP address of the machine that is running the server. The default is ``None``,
        in which case localhost is used.
    port_num : int, default: 50051
        Port number that the server is listening on.
    """
    MOD.current_session = _Session(ip_address, port_num, None)
    MOD.current_session.connect()
    return MOD.current_session


def launch_session(ansys_em_root, port_num=None):
    r"""Launch a local session to an EDB API server.

    The session must be manually disconnected after use by calling session.disconnect()

    Parameters
    ----------
    ansys_em_root : str
        Directory where the ``EDB_RPC_Server.exe`` file is installed.
    port_num : int, default: None
        Port number to listen on. The default is ``None``, in which
        case localhost is used.

    Examples
    --------
    Create a session and then disconnect it

    >>> session = launch_session("C:\\Program Files\\AnsysEM\\v231\\Win64", 50051)
    >>> # program goes here
    >>> session.disconnect()
    """
    ip_address = None  # remote launch is not supported yet

    try:
        _ensure_session(ansys_em_root, port_num, ip_address)
        return MOD.current_session
    except Exception as e:  # noqa
        if MOD.current_session is not None:
            MOD.current_session.disconnect()
        raise


@contextmanager
def session(ansys_em_root, port_num, ip_address=None):
    r"""Launch a local session to an EDB API server in a context manager.

    Parameters
    ----------
    ansys_em_root : str
        Directory where the ``EDB_RPC_Server.exe`` file is installed.
    port_num : int
        Port number to listen on.
    ip_address : str, default: None
        IP address where the server executable file is running. The default is ``None``, in which
        case localhost is used.

        .. note::
           This parameter is currently not supported. In future releases, this parameter is to
           support remotely running the API on another machine.

    Examples
    --------
    Create a session that automatically disconnects when it goes out of scope.

    >>> with session("C:\\Program Files\\AnsysEM\\v231\\Win64", 50051):
    >>>    # program goes here
    """
    try:
        _ensure_session(ansys_em_root, port_num, ip_address)
        yield
    except EDBSessionException:
        raise
    except Exception as e:  # noqa
        raise
    finally:
        MOD.current_session.disconnect()


def get_layer_collection_stub():
    """Get the layer collection stub.

    Returns
    -------
    LayerCollectionServiceStub
    """
    return StubAccessor(StubType.layer_collection).__get__()


def get_stackup_layer_stub():
    """Get the stackup layer stub.

    Returns
    -------
    StackupLayerServiceStub
    """
    return StubAccessor(StubType.stackup_layer).__get__()


def get_via_layer_stub():
    """Get the via layer stub.

    Returns
    -------
    ViaLayerServiceStub
    """
    return StubAccessor(StubType.via_layer).__get__()


def get_variable_server_stub():
    """Get the variable server stub.

    Returns
    -------
    VariableServerServiceStub
    """
    return StubAccessor(StubType.variable_server).__get__()


def _ensure_session(ansys_em_root, port_num, ip_address):
    """Check for a running local session and create one if it doesn't exist.

    Parameters
    ----------
    ansys_em_root : str
        Directory where the ``EDB_RPC_Server.exe`` file is installed.
    port_num : int
        Port number to listen on.
    ip_address : str, default: None
        IP address where the server executable file is running.
    """
    if MOD.current_session is not None:
        if MOD.current_session.port_num != port_num:
            raise EDBSessionException(ErrorCode.STARTUP_MULTI_SESSIONS)
    else:
        MOD.current_session = _Session(ip_address, port_num, ansys_em_root)
        MOD.current_session.connect()



"""This module performs conversions from arbitrary user input to explicit types."""
from ansys.api.edb.v1.edb_messages_pb2 import ValueMessage

from ansys.edb.core.geometry import point3d_data, point_data
from ansys.edb.core.utility import value


def to_value(val):
    """Take a value implicitly convertible to a ``Value``type and return as a ``Value`` type.

    Parameters
    ----------
    val : ansys.edb.core.typing.ValueLike

    Returns
    -------
    :class:`.Value`
    """
    if isinstance(val, value.Value):
        return val
    elif type(val) in [int, float, complex, str, ValueMessage]:
        return value.Value(val)
    else:
        raise TypeError(
            f"Value-like objects must be either of type Value or int/float/complex/str. - Received '{val}'"
        )


def to_point(val):
    """Take a value implicitly convertible to a ``PointData`` type and return as a ``PointData`` type.

    Parameters
    ----------
    val : ansys.edb.core.typing.PointLike

    Returns
    -------
    :class:`.PointData`
    """
    if isinstance(val, point_data.PointData):
        return val
    try:
        if len(val) == 2:
            return point_data.PointData(val)
    except TypeError:
        return point_data.PointData(val)

    raise TypeError(
        "Point-like objects must be either of type PointData or a list/tuple containing (start, end) or (arc_height)."
    )


def to_point3d(val):
    """Convert a value to a ``Point3DData`` object.

    Parameters
    ----------
    val : Point3DData, tuple[:term:`ValueLike`,:term:`ValueLike`,:term:`ValueLike`]
        Value to convert.

    Returns
    -------
    :class:`.Point3DData`
    """
    if isinstance(val, point3d_data.Point3DData):
        return val
    if len(val) == 3:
        return point3d_data.Point3DData(val[0], val[1], val[2])



"""IC component property."""

from ansys.api.edb.v1.ic_component_property_pb2_grpc import ICComponentPropertyServiceStub
import google.protobuf.empty_pb2 as empty_pb2

from ansys.edb.core.definition import (
    component_property,
    die_property,
    port_property,
    solder_ball_property,
)
from ansys.edb.core.inner import messages
from ansys.edb.core.session import StubAccessor, StubType


class ICComponentProperty(component_property.ComponentProperty):
    """Represents an IC component property."""

    __stub: ICComponentPropertyServiceStub = StubAccessor(StubType.ic_component_property)

    @classmethod
    def create(cls):
        """
        Create an IC component property.

        Returns
        -------
        ICComponentProperty
            IC component property created.
        """
        return ICComponentProperty(cls.__stub.Create(empty_pb2.Empty()))

    @property
    def solder_ball_property(self):
        """:obj:`SolderBallProperty`: Solder ball property.

        A copy is returned. Use the setter for any modifications to be reflected.
        """
        return solder_ball_property.SolderBallProperty(
            self.__stub.GetSolderBallProperty(messages.edb_obj_message(self))
        )

    @solder_ball_property.setter
    def solder_ball_property(self, value):
        self.__stub.SetSolderBallProperty(
            messages.pointer_property_message(target=self, value=value)
        )

    @property
    def die_property(self):
        """:obj:`DieProperty`: Die property.

        A copy is returned. Use the setter for any modifications to be reflected.
        """
        return die_property.DieProperty(self.__stub.GetDieProperty(messages.edb_obj_message(self)))

    @die_property.setter
    def die_property(self, value):
        self.__stub.SetDieProperty(messages.pointer_property_message(target=self, value=value))

    @property
    def port_property(self):
        """:obj:`PortProperty`: Port property.

        A copy is returned. Use the setter for any modifications to be reflected.
        """
        return port_property.PortProperty(
            self.__stub.GetPortProperty(messages.edb_obj_message(self))
        )

    @port_property.setter
    def port_property(self, value):
        self.__stub.SetPortProperty(messages.pointer_property_message(target=self, value=value))



"""This module allows for the creating of objects while avoid circular imports."""

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.hierarchy import cell_instance
from ansys.edb.core.hierarchy.group import Group
from ansys.edb.core.hierarchy.pin_group import PinGroup
from ansys.edb.core.layout import voltage_regulator
from ansys.edb.core.primitive.primitive import PadstackInstance, Primitive
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.terminal.terminals import Terminal, TerminalInstance


def create_conn_obj(msg):
    """Create a connection object of its derived type based on its layout object type.

    Parameters
    ----------
    msg : ansys.api.edb.v1.edb_messages_pb2.EDBObjMessage

    Returns
    -------
    ansys.edb.core.inner.ConnObj
    """
    type = LayoutObjType(StubAccessor(StubType.connectable).__get__().GetObjType(msg).type)
    if type == LayoutObjType.PRIMITIVE:
        return Primitive(msg).cast()
    elif type == LayoutObjType.PADSTACK_INSTANCE:
        return PadstackInstance(msg)
    elif type == LayoutObjType.TERMINAL:
        return Terminal(msg).cast()
    elif type == LayoutObjType.TERMINAL_INSTANCE:
        return TerminalInstance(msg)
    elif type == LayoutObjType.CELL_INSTANCE:
        return cell_instance.CellInstance(msg)
    elif type == LayoutObjType.GROUP:
        return Group(msg).cast()
    elif type == LayoutObjType.PIN_GROUP:
        return PinGroup(msg)
    elif type == LayoutObjType.VOLTAGE_REGULATOR:
        return voltage_regulator.VoltageRegulator(msg)
    raise TypeError("Encountered an unknown layout object type.")



"""Component model definition."""
from ansys.api.edb.v1.component_model_pb2_grpc import (
    ComponentModelServiceStub,
    DynamicLinkComponentModelServiceStub,
    NPortComponentModelServiceStub,
)
import google.protobuf.wrappers_pb2 as proto_wrappers

from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.session import StubAccessor, StubType


class ComponentModel(ObjBase):
    """Represents a component model."""

    __stub: ComponentModelServiceStub = StubAccessor(StubType.component_model)

    @property
    def reference_file(self):
        """:obj:`str`: Name of the reference file associated with the component model."""
        return self.__stub.GetReferenceFile(self.msg).value

    @reference_file.setter
    def reference_file(self, value):
        self.__stub.SetReferenceFile(messages.string_property_message(self, value))

    @classmethod
    def find_by_name(cls, comp_def, value):
        """Find a component model by name in a given component def.

        Parameters
        ----------
        comp_def : :class:`ComponentDef <ansys.edb.core.definition.ComponentDef>`
            Component def to search for the component model.
        value : str
            Name of the component model.

        Returns
        -------
        ComponentModel
            Component model that is found, ``None`` otherwise.
        """
        return ComponentModel(
            cls.__stub.FindByName(messages.string_property_message(comp_def, value))
        )

    @classmethod
    def find_by_id(cls, comp_def, value):
        """Find a component model by ID in a given component def.

        Parameters
        ----------
        comp_def : :class:`ComponentDef <ansys.edb.core.definition.ComponentDef>`
            Component def to search for the component model.
        value : int
            ID of the component model.

        Returns
        -------
        ComponentModel
            Component model that is found, ``None`` otherwise.
        """
        return ComponentModel(cls.__stub.FindById(messages.int_proprty_message(comp_def, value)))


class NPortComponentModel(ComponentModel):
    """Represents an NPort component model."""

    __stub: NPortComponentModelServiceStub = StubAccessor(StubType.nport_component_model)

    @classmethod
    def create(cls, name):
        """Create an NPort component model.

        Parameters
        ----------
        name : str
            Name of the NPport component model.

        Returns
        -------
        NPortComponentModel
            NPort component model created.

        Notes
        -----
        The component model does not belong to a specific database until it is added to a
        :class:`.ComponentDef` instance.
        """
        return NPortComponentModel(cls.__stub.Create(proto_wrappers.StringValue(value=name)))


class DynamicLinkComponentModel(ComponentModel):
    """Represents a dynamic link component model."""

    __stub: DynamicLinkComponentModelServiceStub = StubAccessor(StubType.dyn_link_component_model)

    @classmethod
    def create(cls, name):
        """Create a dynamic link component model.

        Parameters
        ----------
        name : str
            Name of the dynamic link component model.

        Returns
        -------
        DynamicLinkComponentModel
            Dynamic link component model created.

        Notes
        -----
        The component model does not belong to a specific database until it is added to a
        :class:`.ComponentDef` class.
        """
        return DynamicLinkComponentModel(cls.__stub.Create(proto_wrappers.StringValue(value=name)))

    @property
    def design_name(self):
        """:obj:`str`: Name of the design associated with the dynamic link component model."""
        return self.__stub.GetDesignName(self.msg).value

    @design_name.setter
    def design_name(self, value):
        self.__stub.SetDesignName(messages.string_property_message(self, value))



"""Client-side gRPC interceptors."""

import abc
import logging

from grpc import StatusCode, UnaryUnaryClientInterceptor

from ansys.edb.core.inner.exceptions import EDBSessionException, ErrorCode, InvalidArgumentException


class Interceptor(UnaryUnaryClientInterceptor, metaclass=abc.ABCMeta):
    """Provides the base interceptor class."""

    def __init__(self, logger):
        """Initialize a base interceptor with a logger."""
        super().__init__()
        self._logger = logger

    @abc.abstractmethod
    def _post_process(self, response):
        pass

    def intercept_unary_unary(self, continuation, client_call_details, request):
        """Intercept a gRPC call."""
        response = continuation(client_call_details, request)

        self._post_process(response)

        return response


class LoggingInterceptor(Interceptor):
    """Logs EDB errors on each request."""

    _LOG_LEVELS = {
        "log-fatal": logging.CRITICAL,
        "log-error": logging.ERROR,
        "log-warn": logging.WARN,
        "log-info": logging.INFO,
    }

    def _handler(self, log_level):
        if log_level == logging.CRITICAL:
            return self._logger.critical
        elif log_level == logging.WARN:
            return self._logger.warning
        elif log_level == logging.ERROR:
            return self._logger.error
        elif log_level == logging.INFO:
            return self._logger.info
        else:
            return self._logger.logger

    def _post_process(self, response):
        tm = response.trailing_metadata()
        if tm is not None:
            for kv in tm:
                if kv[0] in self._LOG_LEVELS:
                    handler = self._handler(self._LOG_LEVELS[kv[0]])
                    handler(kv[1])


class ExceptionInterceptor(Interceptor):
    """Handles general gRPC errors on each request."""

    def _post_process(self, response):
        code = response.code()
        exception = None

        if code == StatusCode.UNAVAILABLE:
            exception = EDBSessionException(ErrorCode.UNAVAILABLE)
            self._logger.critical(repr(exception))
        elif code == StatusCode.INVALID_ARGUMENT:
            exception = InvalidArgumentException(response)
            self._logger.error(repr(exception))

        if exception is not None:
            raise exception



"""Dataset definition."""
from ansys.api.edb.v1.dataset_def_pb2_grpc import DatasetDefServiceStub

from ansys.edb.core.edb_defs import DefinitionObjType
from ansys.edb.core.inner import ObjBase
from ansys.edb.core.inner.messages import (
    edb_obj_message,
    edb_obj_name_message,
    points_property_message,
    string_property_message,
)
from ansys.edb.core.inner.parser import to_point_data_list
from ansys.edb.core.session import StubAccessor, StubType


class DatasetDef(ObjBase):
    """Represents a dataset definition."""

    __stub: DatasetDefServiceStub = StubAccessor(StubType.dataset_def)

    @classmethod
    def create(cls, database, name):
        """Create a dataset definition object.

        Parameters
        ----------
        database : :class:`.Database`
            Database to create the dataset definition in.
        name : :obj:`str`
            Name of the dataset to create.


        Returns
        -------
        DatasetDef
        """
        return DatasetDef(cls.__stub.Create(string_property_message(database, name)))

    @classmethod
    def find_by_name(cls, database, name):
        """Find a dataset definition by name in a given database.

        Parameters
        ----------
        database : :class:`.Database`
            Database to search for the dataset definition.
        name : :obj:`str`
            Name of the dataset definition.

        Returns
        -------
        DatasetDef
            Dataset definition object found.
            If a dataset isn't found, the dataset's ``is_null`` attribute is set to ``True``.
        """
        return DatasetDef(cls.__stub.FindByName(edb_obj_name_message(database, name)))

    @property
    def definition_type(self):
        """:class:`.DefinitionObjType`: Definition type."""
        return DefinitionObjType.DATASET_DEF

    @property
    def name(self):
        """:obj:`str`: Name of the dataset definition."""
        return self.__stub.GetName(edb_obj_message(self)).value

    @name.setter
    def name(self, name):
        self.__stub.SetName(string_property_message(self, name))

    @to_point_data_list
    def get_data(self):
        """Get a list of data points in the dataset definition.

        Returns
        -------
        list[:class:`.PointData`]
        """
        msg = self.__stub.GetData(edb_obj_message(self))
        return msg.points

    def set_data(self, points):
        """Set a list of data points in the dataset definition.

        Parameters
        ----------
        points : list[:class:`.PointData`]
        """
        self.__stub.SetData(points_property_message(self, points))



"""Triangle 3D data."""


class Triangle3DData:
    """Represents a triangle defined by three 3D points."""

    def __init__(self, point_1, point_2, point_3):
        """Create a 3D triangle.

        Parameters
        ----------
        point_1: :class:`Point3DData`
        point_2: :class:`Point3DData`
        point_3: :class:`Point3DData`
        """
        self._point_1 = point_1
        self._point_2 = point_2
        self._point_3 = point_3

    @property
    def point_1(self):
        """:class:`Point3DData`: First 3D point."""
        return self._point_1

    @point_1.setter
    def point_1(self, point_1):
        self._point_1 = point_1

    @property
    def point_2(self):
        """:class:`Point3DData`: Second 3D point."""
        return self._point_2

    @point_2.setter
    def point_2(self, point_2):
        self._point_2 = point_2

    @property
    def point_3(self):
        """:class:`Point3DData`: Third 3D point."""
        return self._point_3

    @point_3.setter
    def point_3(self, point_3):
        self._point_3 = point_3



"""Die property."""

from enum import Enum

import ansys.api.edb.v1.die_property_pb2 as die_property_pb2
from ansys.api.edb.v1.die_property_pb2_grpc import DiePropertyServiceStub
import google.protobuf.empty_pb2 as empty_pb2

from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class DieOrientation(Enum):
    """Provides an enum representing die orientations."""

    CHIP_UP = die_property_pb2.DIE_ORIENTATION_CHIP_UP
    CHIP_DOWN = die_property_pb2.DIE_ORIENTATION_CHIP_DOWN


class DieType(Enum):
    """Provides an enum representing die types."""

    NONE = die_property_pb2.DIE_TYPE_NONE
    FLIPCHIP = die_property_pb2.DIE_TYPE_FLIPCHIP
    WIREBOND = die_property_pb2.DIE_TYPE_WIREBOND


class DieProperty(ObjBase):
    """Represents a die property."""

    __stub: DiePropertyServiceStub = StubAccessor(StubType.die_property)

    @classmethod
    def create(cls):
        """
        Create a die property.

        Returns
        -------
        DieProperty
            Die property created.
        """
        return DieProperty(cls.__stub.Create(empty_pb2.Empty()))

    def clone(self):
        """
        Clone a die property.

        Returns
        -------
        DieProperty
            Die property cloned.
        """
        return DieProperty(self.__stub.Clone(messages.edb_obj_message(self)))

    @property
    def die_type(self):
        """:obj:`DieType`: Die type."""
        return DieType(self.__stub.GetDieType(self.msg).die_type)

    @die_type.setter
    def die_type(self, value):
        self.__stub.SetDieType(messages.set_die_type_message(self, value))

    @property
    def height(self):
        """:class:`.Value`: Die height.

        This property can be set with :term:`ValueLike`.
        """
        return Value(self.__stub.GetHeight(messages.edb_obj_message(self)))

    @height.setter
    def height(self, height):
        self.__stub.SetHeight(messages.value_property_message(self, messages.value_message(height)))

    @property
    def die_orientation(self):
        """:obj:`DieOrientation`: Die orientation."""
        return DieOrientation(self.__stub.GetOrientation(self.msg).die_orientation)

    @die_orientation.setter
    def die_orientation(self, value):
        self.__stub.SetOrientation(messages.set_die_orientation_message(self, value))



"""Terminals."""

from enum import Enum

import ansys.api.edb.v1.edge_term_pb2 as edge_term_pb2
import ansys.api.edb.v1.term_pb2 as term_pb2

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.geometry.arc_data import ArcData
from ansys.edb.core.inner import ObjBase, TypeField, conn_obj, messages, parser
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.port_post_processing_prop import PortPostProcessingProp
from ansys.edb.core.utility.value import Value


class TerminalType(Enum):
    """Provides an enum representing terminal types."""

    EDGE = term_pb2.EDGE_TERM
    POINT = term_pb2.POINT_TERM
    TERM_INST = term_pb2.TERM_INST_TERM
    PADSTACK_INST = term_pb2.PADSTACK_INST_TERM
    BUNDLE = term_pb2.BUNDLE_TERM
    PIN_GROUP = term_pb2.PIN_GROUP_TERM


class BoundaryType(Enum):
    """Provides an enum representing terminal boundary types."""

    PORT = term_pb2.PORT_BOUNDARY
    PEC = term_pb2.PEC_BOUNDARY
    RLC = term_pb2.RLC_BOUNDARY
    CURRENT_SOURCE = term_pb2.CURRENT_SOURCE
    VOLTAGE_SOURCE = term_pb2.VOLTAGE_SOURCE
    NEXXIM_GROUND = term_pb2.NEXXIM_GROUND
    NEXXIM_PORT = term_pb2.NEXXIM_PORT
    DC_TERMINAL = term_pb2.DC_TERMINAL
    VOLTAGE_PROBE = term_pb2.VOLTAGE_PROBE


class SourceTermToGroundType(Enum):
    """Provides an enum representing source terminal-to-ground types."""

    NO_GROUND = term_pb2.NO_GROUND
    NEGATIVE = term_pb2.NEGATIVE
    POSITIVE = term_pb2.POSITIVE


class HfssPIType(Enum):
    """Provides an enum representing HFSS PI types."""

    DEFAULT = term_pb2.PI_DEFAULT
    COAXIAL_OPEN = term_pb2.PI_COAXIAL_OPEN
    COAXIAL_SHORTENED = term_pb2.PI_COAXIAL_SHORTENED
    GAP = term_pb2.PI_GAP
    LUMPED = term_pb2.PI_LUMPED


class EdgeType(Enum):
    """Provides an enum representing edge types."""

    PRIMITIVE = edge_term_pb2.PRIMITIVE_EDGE
    PADSTACK = edge_term_pb2.PAD_EDGE


class Edge(ObjBase):
    """Represents an edge."""

    __stub = StubAccessor(StubType.edge)
    type = TypeField(None)

    def cast(self):
        """Cast the base edge object to the correct subclass, if possible.

        Returns
        -------
        Edge
        """
        if self.is_null:
            return

        tp = self.type
        if tp == EdgeType.PRIMITIVE:
            return PrimitiveEdge(self.msg)
        if tp == EdgeType.PADSTACK:
            return PadEdge(self.msg)

    @classmethod
    def _create(cls, **params):
        return cls.__stub.Create(messages.edge_creation_message(cls.type.value, **params))

    @property
    def _type(self):
        return EdgeType(self.__stub.GetType(self.msg))

    @property
    def _params(self):
        res = self.__stub.GetParameters(self.msg)
        if self.type == EdgeType.PRIMITIVE:
            return res.primitive_params
        elif self.type == EdgeType.PADSTACK:
            return res.pad_params


class PadEdge(Edge):
    """Represents a padstack edge."""

    type = TypeField(EdgeType.PADSTACK)

    @classmethod
    def create(cls, padstack_instance, layer, arc):
        """Create a padstack edge.

        Parameters
        ----------
        padstack_instance : :class:`.PadstackInstance`
        layer : :class:`.Layer` or :obj:`str`
        arc : :class:`.ArcData`

        Returns
        -------
        PadEdge
        """
        return PadEdge(cls._create(padstack_instance=padstack_instance, layer=layer, arc=arc))

    @property
    def padstack_instance(self):
        """:class:`.PadstackInstance`: Padstack instance that the edge is on."""
        from ansys.edb.core.primitive import primitive

        return primitive.PadstackInstance(self._params.padstack_instance)

    @property
    def layer(self):
        """:class:`.Layer`: Layer that the edge is on."""
        return Layer(self._params.layer.id).cast()

    @property
    def arc(self):
        """:class:`.ArcData`: Arc of the edge."""
        return ArcData(self._params.arc)


class PrimitiveEdge(Edge):
    """Represents a primitive edge."""

    type = TypeField(EdgeType.PRIMITIVE)

    @classmethod
    def create(cls, prim, point):
        """Create a primitive edge.

        Parameters
        ----------
        prim : :class:`.Primitive`
        point : :term:`Point2DLike`

        Returns
        -------
        PrimitiveEdge
        """
        return PrimitiveEdge(cls._create(primitive=prim, point=point))

    @property
    def primitive(self):
        """:class:`.Primitive`: Primitive of the terminal."""
        from ansys.edb.core.primitive import primitive

        return primitive.Primitive(self._params.primitive).cast()

    @property
    @parser.to_point_data
    def point(self):
        """:class:`.PointData`: Coordinates (x, y) of the terminal."""
        return self._params.point


class Terminal(conn_obj.ConnObj):
    """Represents a terminal object."""

    __stub = StubAccessor(StubType.terminal)
    type = TypeField(None)
    layout_obj_type = LayoutObjType.TERMINAL

    def cast(self, term_type=None):
        """Cast the terminal object to the correct concrete type, fetching the type if necessary.

        Parameters
        ----------
        term_type : TerminalType

        Returns
        -------
        Terminal
        """
        if self.is_null:
            return

        tt = self.type

        if term_type is not None and tt != term_type:
            return
        elif tt == TerminalType.EDGE:
            return EdgeTerminal(self.msg)
        elif tt == TerminalType.POINT:
            return PointTerminal(self.msg)
        elif tt == TerminalType.TERM_INST:
            return TerminalInstanceTerminal(self.msg)
        elif tt == TerminalType.PADSTACK_INST:
            return PadstackInstanceTerminal(self.msg)
        elif tt == TerminalType.BUNDLE:
            return BundleTerminal(self.msg)
        elif tt == TerminalType.PIN_GROUP:
            return PinGroupTerminal(self.msg)

    @classmethod
    def find(cls, layout, name):
        """Find a terminal by name in a given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
           Layout to search for the terminal.
        name : .str
            Name of the terminal.

        Returns
        -------
        Terminal
        """
        return Terminal(
            cls.__stub.FindByName(messages.term_find_by_name_message(layout, name))
        ).cast(cls.type)

    @property
    def _params(self):
        return self.__stub.GetParams(self.msg)

    @_params.setter
    def _params(self, values):
        self.__stub.SetParams(messages.term_set_params_message(self, **values))

    @property
    def _type(self):
        """:class:`TerminalType`: Terminal type."""
        return TerminalType(self._params.term_type)

    @property
    def bundle_terminal(self):
        """:class:`BundleTerminal`: Bundle terminal that the terminal belongs to, if any."""
        return BundleTerminal(self._params.bundle_term)

    @property
    def reference_terminal(self):
        """:class:`Terminal`: Terminal that the terminal references, if any."""
        return Terminal(self._params.ref_term).cast()

    @reference_terminal.setter
    def reference_terminal(self, value):
        self._params = {"ref_term": value}

    @property
    def reference_layer(self):
        """:class:`.Layer`: Layer that the terminal references, if any, by either layer object or name."""
        return Layer(self._params.ref_layer).cast()

    @reference_layer.setter
    def reference_layer(self, value):
        self._params = {"ref_layer": value}

    @property
    def is_interface_terminal(self):
        """:obj:`bool`: Flag indicating if the terminal is an interface."""
        return self._params.is_interface

    @property
    def is_reference_terminal(self):
        """:obj:`bool`: Flag indicating if the terminal is a reference terminal."""
        return self._params.is_reference

    @property
    def use_reference_from_hierarchy(self):
        """:obj:`bool`: Flag indicating if the terminal can use references from the hierarchy."""
        return self._params.use_ref_from_hierarchy

    @use_reference_from_hierarchy.setter
    def use_reference_from_hierarchy(self, value):
        self._params = {"use_ref_from_hierarchy": value}

    @property
    def is_auto_port(self):
        """:obj:`bool`: Flag indicating if the terminal is an auto port."""
        return self._params.is_auto_port

    @is_auto_port.setter
    def is_auto_port(self, value):
        self._params = {"is_auto_port": value}

    @property
    def is_circuit_port(self):
        """:obj:`bool`: Flag indicating if the terminal is a circuit port."""
        return self._params.is_circuit_port

    @is_circuit_port.setter
    def is_circuit_port(self, value):
        self._params = {"is_circuit_port": value}

    @property
    def name(self):
        """:obj:`str`: Name of the terminal."""
        return self._params.name

    @name.setter
    def name(self, value):
        self._params = {"name": value}

    @property
    def boundary_type(self):
        """:class:`BoundaryType`: Boundary type of the terminal."""
        return BoundaryType(self._params.boundary_type)

    @boundary_type.setter
    def boundary_type(self, value):
        self._params = {"boundary_type": value.value}

    @property
    def impedance(self):
        """:class:`.Value`: Impedance of the terminal."""
        return Value(self._params.impedance)

    @impedance.setter
    def impedance(self, value):
        self._params = {"impedance": value}

    @property
    def source_amplitude(self):
        """:class:`.Value`: Source amplitude of the terminal."""
        return Value(self._params.source_amplitude)

    @source_amplitude.setter
    def source_amplitude(self, value):
        self._params = {"source_amplitude": value}

    @property
    def source_phase(self):
        """:class:`.Value`: Source phase of the terminal."""
        return Value(self._params.source_phase)

    @source_phase.setter
    def source_phase(self, value):
        self._params = {"source_phase": value}

    @property
    def term_to_ground(self):
        """:class:`SourceTermToGroundType`: Source terminal-to-ground type."""
        return SourceTermToGroundType(self._params.term_to_ground)

    @term_to_ground.setter
    def term_to_ground(self, value):
        self._params = {"term_to_ground": value.value}

    @property
    def hfss_pi_type(self):
        """:class:`HfssPIType`: HFSS PI type."""
        return HfssPIType(self._params.hfss_pi_type)

    @hfss_pi_type.setter
    def hfss_pi_type(self, value):
        self._params = {"hfss_pi_type": value.value}

    @property
    def model(self):
        """:obj:`str`: S-parameter model."""
        return self._params.s_param_model

    @model.setter
    def model(self, value):
        self._params = {"s_param_model": value}

    @property
    @parser.to_rlc
    def rlc_boundary_parameters(self):
        """:class:`.Rlc`: RLC boundary parameters."""
        return self._params.rlc

    @rlc_boundary_parameters.setter
    def rlc_boundary_parameters(self, value):
        self._params = {"rlc": value}

    @property
    def port_post_processing_prop(self):
        """:class:`.PortPostProcessingProp`: Port postprocessing properties."""
        msg = self._params.port_post_processing_prop
        return PortPostProcessingProp(
            voltage_magnitude=msg.voltage_magnitude,
            voltage_phase=msg.voltage_phase,
            deembed_length=msg.deembed_length,
            renormalization_impedance=msg.renormalization_impedance,
            do_deembed=msg.do_deembed,
            do_deembed_gap_l=msg.do_deembed_gap_length,
            do_renormalize=msg.do_renormalize,
        )

    @port_post_processing_prop.setter
    def port_post_processing_prop(self, value):
        self._params = {"port_post_processing_prop": value}

    def _product_solvers(self, product_id):
        return self.__stub.GetProductSolvers(
            messages.term_get_product_solver_message(self, product_id)
        )

    def product_solver_option(self, product_id, solver_name):
        """Get the name of the product solver option.

        Parameters
        ----------
        product_id : ProductIdType
            ID of the product.
        solver_name : :obj:`str`
            Name of the solver.

        Returns
        -------
        :obj:`str`
            Name of the product solver option.
        """
        return next(
            solver.option
            for solver in self._product_solvers(product_id)
            if solver.name == solver_name
        )

    def set_product_solver_option(self, product_id, solver_name, option):
        """Set the product solver option.

        Parameters
        ----------
        product_id : ProductIdType
            ID of the product.
        solver_name : obj:`str`
            Name of the solver.
        option : obj:`str`
           Name of the product solver option.
        """
        self.__stub.SetProductSolverOptions(
            messages.term_set_solver_option_message(self, product_id, solver_name, option)
        )

    def product_solver_names(self, product_id):
        """Get the list of solver names.

        Parameters
        ----------
        product_id : ProductIdType
            ID of the product.

        Returns
        -------
        list of str
        """
        return [solver.name for solver in self._product_solvers(product_id)]


class TerminalInstance(conn_obj.ConnObj):
    """Represents a terminal instance."""

    __stub = StubAccessor(StubType.terminal_instance)
    layout_obj_type = LayoutObjType.TERMINAL_INSTANCE

    @classmethod
    def create(cls, layout, cell_instance, name, net_ref):
        """Create a terminal instance.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the terminal instance in.
        cell_instance : :class:`.CellInstance`
            Name of the cell instance to create the terminal instance on.
        name : :obj:`str`
            Name of the terminal instance.
        net_ref : :class:`.Net` or :obj:`str` or None
            Net reference.

        Returns
        -------
        TerminalInstance
        """
        return TerminalInstance(
            cls.__stub.Create(
                messages.term_inst_creation_message(layout, net_ref, cell_instance, name)
            )
        )

    @property
    def owning_cell_instance(self):
        """:class:`.CellInstance`: Cell instance that owns the terminal."""
        from ansys.edb.core.hierarchy import cell_instance

        return cell_instance.CellInstance(self.__stub.GetOwningCellInstance(self.msg))

    @property
    def definition_terminal(self):
        """:class:`Terminal`: Definition terminal, if any."""
        return Terminal(self.__stub.GetDefinitionTerminal(self.msg)).cast()

    @property
    def definition_terminal_name(self):
        """:obj:`str`: Name of the definition terminal."""
        return self.__stub.GetDefinitionTerminalName(self.msg).value


class TerminalInstanceTerminal(Terminal):
    """Represents a terminal instance terminal."""

    __stub = StubAccessor(StubType.terminal_instance_terminal)
    type = TypeField(TerminalType.TERM_INST)

    @classmethod
    def create(cls, layout, term_instance, name, net_ref, is_ref=False):
        """Create a terminal instance terminal.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the terminal instance terminal in.
        term_instance : TerminalInstance
            Terminal instance to create the terminal instance terminal on.
        name : :obj:`str`
            Name of the terminal instance terminal.
        net_ref : :class:`.Net` or :obj:`str` or None
            Net reference.
        is_ref : bool, default: False
            Whether the terminal instance terminal is a reference terminal.

        Returns
        -------
        TerminalInstanceTerminal
        """
        return TerminalInstanceTerminal(
            cls.__stub.Create(
                messages.term_inst_term_creation_message(
                    layout, net_ref, name, term_instance, is_ref
                )
            )
        )

    @property
    def terminal_instance(self):
        """:class:`TerminalInstance`: Terminal instance."""
        return TerminalInstance(self.__stub.GetTerminalInstance(self.msg))

    @terminal_instance.setter
    def terminal_instance(self, value):
        self.__stub.SetTerminalInstance(messages.term_inst_term_set_instance_message(self, value))


class BundleTerminal(Terminal):
    """Represents a bundle terminal object."""

    __stub = StubAccessor(StubType.bundle_terminal)
    type = TypeField(TerminalType.BUNDLE)

    @classmethod
    def create(cls, terminals):
        """Create a bundle terminal.

        Parameters
        ----------
        terminals : list of Terminal

        Returns
        -------
        BundleTerminal
        """
        return BundleTerminal(cls.__stub.Create(messages.edb_obj_collection_message(terminals)))

    @property
    def terminals(self):
        """:obj:`list` of Terminal: All terminals grouped in the terminal."""
        return [Terminal(msg).cast() for msg in self.__stub.GetTerminals(self.msg)]

    def ungroup(self):
        """Delete the grouping."""
        self.__stub.Ungroup(self.msg)
        self.msg = None


class PointTerminal(Terminal):
    """Represents a point terminal object."""

    __stub = StubAccessor(StubType.point_terminal)
    type = TypeField(TerminalType.POINT)

    @classmethod
    def create(cls, layout, net, layer, name, point):
        """Create a point terminal.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the point terminal in.
        net : :class:`.Net` or :obj:`str` or None
            Net.
        layer : :class:`.Layer` or :obj:`str`
            Layer to place the point terminal on.
        name : :obj:`str`
            Name of the point terminal.
        point : :term:`Point2DLike`
            Type of the point terminal.

        Returns
        -------
        PointTerminal
        """
        return PointTerminal(
            cls.__stub.Create(messages.point_term_creation_message(layout, net, layer, name, point))
        )

    @property
    def params(self):
        """:class:`.Layer`, :class:`.PointData`: Layer that the point terminal is placed on and \
        the (x, y) coordinates."""
        res = self.__stub.GetParameters(self.msg)
        point = parser.to_point_data(res.point)
        layer = Layer(res.layer.id).cast()
        return layer, point

    @params.setter
    def params(self, params):
        self.__stub.SetParameters(messages.point_term_set_params_message(self, *params))

    @property
    def layer(self):
        """:class:`.Layer`: Layer that the point terminal is placed on."""
        return self.params[0]

    @layer.setter
    def layer(self, value):
        self.params = (messages.edb_obj_message(value), self.point)

    @property
    def point(self):
        """:class:`.PointData`: Coordinates (x, y) of the point terminal."""
        return self.params[1]

    @point.setter
    def point(self, value):
        self.params = (self.layer, messages.point_message(value))


class PadstackInstanceTerminal(Terminal):
    """Represents a padstack instance terminal."""

    __stub = StubAccessor(StubType.padstack_instance_terminal)
    type = TypeField(TerminalType.PADSTACK_INST)

    @classmethod
    def create(cls, layout, name, padstack_instance, layer, net, is_ref=False):
        """Create a padstack instance terminal.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the padstack instance terminal in.
        name : :obj:`str`
            Name of the padstack instance terminal.
        padstack_instance : :class:`.PadstackInstance`
            Padstack instance.
        layer : :class:`.Layer` or :obj:`str`
            Layer to place the padstack instance terminal on.
        net : :class:`.Net` or :obj:`str` or None
            Net.
        is_ref : :obj:`bool`, default: False
            Whether the padstack instance terminal is a reference terminal.

        Returns
        -------
        PadstackInstanceTerminal
        """
        return PadstackInstanceTerminal(
            cls.__stub.Create(
                messages.padstack_inst_term_creation_message(
                    layout, name, padstack_instance, layer, net, is_ref
                )
            )
        )

    @property
    def params(self):
        """:obj:`list` of :class:`.PadstackInstance` and :class:`.Layer`: Padstack instance and layer."""
        from ansys.edb.core.primitive import primitive

        res = self.__stub.GetParameters(self.msg)
        padstack_instance = primitive.PadstackInstance(res.padstack_instance)
        layer = Layer(res.layer).cast()
        return padstack_instance, layer

    @params.setter
    def params(self, params):
        (padstack_instance, layer) = params
        self.__stub.SetParameters(
            messages.padstack_inst_term_set_params_message(self, padstack_instance, layer)
        )

    @property
    def padstack_instance(self):
        """:class:`.PadstackInstance`: Padstack instance of the terminal."""
        return self.params[0]

    @property
    def layer(self):
        """:class:`.Layer`: Layer the terminal is placed on."""
        return self.params[1]


class PinGroupTerminal(Terminal):
    """Represents a pin group terminal."""

    __stub = StubAccessor(StubType.pin_group_terminal)
    type = TypeField(TerminalType.PIN_GROUP)

    @classmethod
    def create(cls, layout, name, pin_group, net, is_ref=False):
        """Create a pin group terminal.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the pin group terminal in.
        net : :class:`.Net` or :obj:`str` or None
            Net reference.
        name : :obj:`str`
            Name of the pin group terminal.
        pin_group : :class:`.PinGroup`
            Pin group.
        is_ref : :obj:`bool`, default: False
            Whether the pin group terminal is a reference terminal.

        Returns
        -------
        PinGroupTerminal
        """
        return PinGroupTerminal(
            cls.__stub.Create(
                messages.pin_group_term_creation_message(layout, net, name, pin_group, is_ref)
            )
        )

    @property
    def pin_group(self):
        """:class:`.PinGroup`: Pin group of the terminal."""
        from ansys.edb.core.hierarchy import pin_group

        return pin_group.PinGroup(self.__stub.GetPinGroup(self.msg))

    @pin_group.setter
    def pin_group(self, value):
        self.__stub.SetPinGroup(messages.pin_group_term_set_pin_group_message(self, value))

    @property
    def layer(self):
        """:class:`.Layer`: Layer."""
        return Layer(self.__stub.GetLayer(self.msg)).cast()

    @layer.setter
    def layer(self, value):
        self.__stub.SetLayer(messages.pin_group_term_set_layer_message(self, value))


class EdgeTerminal(Terminal):
    """Represents an edge terminal."""

    __stub = StubAccessor(StubType.edge_terminal)
    type = TypeField(TerminalType.EDGE)

    @classmethod
    def create(cls, layout, name, edges, net=None, is_ref=False):
        """Create an edge terminal.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the edge terminal in.
        name : :obj:`str`
            Name of the edge terminal.
        edges : list of Edge
        net : :class:`.Net` or :obj:`str` or None
            Net reference. The default is ``None``.
        is_ref : :obj:`bool`, default: False
            Whether the edge terminal is a reference terminal.

        Returns
        -------
        EdgeTerminal
        """
        return EdgeTerminal(
            cls.__stub.Create(messages.edge_term_creation_message(layout, net, name, edges, is_ref))
        )

    @property
    def edges(self):
        """:obj:`list` of :class:`.Edge`: All edges on the terminal."""
        return [Edge(msg).cast() for msg in self.__stub.GetEdges(self.msg)]

    @edges.setter
    def edges(self, edges):
        self.__stub.GetEdges(messages.edge_term_set_edges_message(self, edges))



"""Primitive classes."""

from enum import Enum

from ansys.api.edb.v1 import (
    board_bend_def_pb2,
    board_bend_def_pb2_grpc,
    bondwire_pb2,
    bondwire_pb2_grpc,
    circle_pb2,
    circle_pb2_grpc,
    padstack_instance_pb2,
    padstack_instance_pb2_grpc,
    path_pb2,
    path_pb2_grpc,
    polygon_pb2,
    polygon_pb2_grpc,
    primitive_pb2,
    primitive_pb2_grpc,
    rectangle_pb2,
    rectangle_pb2_grpc,
    text_pb2,
    text_pb2_grpc,
)

from ansys.edb.core.definition.padstack_def import PadstackDef
from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import conn_obj, messages, parser
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.layer_map import LayerMap
from ansys.edb.core.utility.value import Value


class PrimitiveType(Enum):
    """Provides an enum representing primitive types."""

    RECTANGLE = primitive_pb2.RECTANGLE
    CIRCLE = primitive_pb2.CIRCLE
    POLYGON = primitive_pb2.POLYGON
    PATH = primitive_pb2.PATH
    BONDWIRE = primitive_pb2.BONDWIRE
    PRIM_PLUGIN = primitive_pb2.PRIM_PLUGIN
    TEXT = primitive_pb2.TEXT
    PATH_3D = primitive_pb2.PATH_3D
    BOARD_BEND = primitive_pb2.BOARD_BEND
    INVALID_TYPE = primitive_pb2.INVALID_TYPE


class RectangleRepresentationType(Enum):
    """Provides an enum representing rectangle types."""

    INVALID_RECT_TYPE = rectangle_pb2.INVALID_RECT_TYPE
    CENTER_WIDTH_HEIGHT = rectangle_pb2.CENTER_WIDTH_HEIGHT
    LOWER_LEFT_UPPER_RIGHT = rectangle_pb2.LOWER_LEFT_UPPER_RIGHT


class PathEndCapType(Enum):
    """Provides an enum representing end cap types."""

    ROUND = path_pb2.ROUND
    FLAT = path_pb2.FLAT
    EXTENDED = path_pb2.EXTENDED
    CLIPPED = path_pb2.CLIPPED
    INVALID = path_pb2.INVALID_END_CAP


class PathCornerType(Enum):
    """Provides an enum representing corner types."""

    ROUND = path_pb2.ROUND_CORNER
    SHARP = path_pb2.SHARP_CORNER
    MITER = path_pb2.MITER_CORNER


class BondwireType(Enum):
    """Provides an enum representing bondwire types."""

    APD = bondwire_pb2.APD_BONDWIRE
    JEDEC4 = bondwire_pb2.JEDEC4_BONDWIRE
    JEDEC5 = bondwire_pb2.JEDEC5_BONDWIRE
    NUM_OF_TYPE = bondwire_pb2.NUM_OF_BONDWIRE_TYPE
    INVALID = bondwire_pb2.INVALID_BONDWIRE_TYPE


class BondwireCrossSectionType(Enum):
    """Provides an enum representing bondwire cross section types."""

    ROUND = bondwire_pb2.BONDWIRE_ROUND
    RECTANGLE = bondwire_pb2.BONDWIRE_RECTANGLE
    INVALID = bondwire_pb2.INVALID_BONDWIRE_CROSS_SECTION_TYPE


class BackDrillType(Enum):
    """Provides an enum representing back drill types."""

    NO_DRILL = padstack_instance_pb2.NO_DRILL
    LAYER_DRILL = padstack_instance_pb2.LAYER_DRILL
    DEPTH_DRILL = padstack_instance_pb2.DEPTH_DRILL


class Primitive(conn_obj.ConnObj):
    """Represents a primitive object."""

    __stub: primitive_pb2_grpc.PrimitiveServiceStub = StubAccessor(StubType.primitive)
    layout_obj_type = LayoutObjType.PRIMITIVE

    def cast(self):
        """Cast the primitive object to the correct concrete type.

        Returns
        -------
        Primitive
        """
        if self.is_null:
            return

        prim_type = self.primitive_type
        if prim_type == PrimitiveType.RECTANGLE:
            return Rectangle(self.msg)
        elif prim_type == PrimitiveType.POLYGON:
            return Polygon(self.msg)
        elif prim_type == PrimitiveType.PATH:
            return Path(self.msg)
        elif prim_type == PrimitiveType.BONDWIRE:
            return Bondwire(self.msg)
        elif prim_type == PrimitiveType.TEXT:
            return Text(self.msg)
        elif prim_type == PrimitiveType.CIRCLE:
            return Circle(self.msg)

    @property
    def primitive_type(self):
        """:class:`PrimitiveType`: Primitive type of the primitive.

        This property is read-only.
        """
        return PrimitiveType(self.__stub.GetPrimitiveType(self.msg).type)

    def add_void(self, hole):
        """Add a void to the primitive.

        Parameters
        ----------
        hole : Primitive
            Void to add.
        """
        self.__stub.AddVoid(
            primitive_pb2.PrimitiveVoidCreationMessage(target=self.msg, hole=hole.msg)
        )

    def set_hfss_prop(self, material, solve_inside):
        """Set HFSS properties.

        Parameters
        ----------
        material : str
            Material property name to set.
        solve_inside : bool
            Whether to solve inside.
        """
        self.__stub.SetHfssProp(
            primitive_pb2.PrimitiveHfssPropMessage(
                target=self.msg, material_name=material, solve_inside=solve_inside
            )
        )

    @property
    def layer(self):
        """:class:`.Layer`: Layer that the primitive object is on."""
        layer_msg = self.__stub.GetLayer(self.msg)
        return Layer(layer_msg).cast()

    @layer.setter
    def layer(self, layer):
        self.__stub.SetLayer(
            primitive_pb2.SetLayerMessage(target=self.msg, layer=messages.layer_ref_message(layer))
        )

    @property
    def is_negative(self):
        """:obj:`bool`: Flag indicating if the primitive is negative."""
        return self.__stub.GetIsNegative(self.msg).value

    @is_negative.setter
    def is_negative(self, is_negative):
        self.__stub.SetIsNegative(
            primitive_pb2.SetIsNegativeMessage(target=self.msg, is_negative=is_negative)
        )

    @property
    def is_void(self):
        """:obj:`bool`: Flag indicating if a primitive is a void."""
        return self.__stub.IsVoid(self.msg).value

    @property
    def has_voids(self):
        """:obj:`bool`: Flag indicating if a primitive has voids inside.

        This property is read-only.
        """
        return self.__stub.HasVoids(self.msg).value

    @property
    def voids(self):
        """:obj:`list` of :class:`.Primitive`: List of void\
        primitive objects inside the primitive.

        This property is read-only.
        """
        return [Primitive(msg).cast() for msg in self.__stub.Voids(self.msg).items]

    @property
    def owner(self):
        """:class:`.Primitive`: Owner of the primitive object.

        This property is read-only.
        """
        return Primitive(self.__stub.GetOwner(self.msg)).cast()

    @property
    def is_parameterized(self):
        """:obj:`bool`: Whether the primitive is parametrized.

        This property is read-only.
        """
        return self.__stub.IsParameterized(self.msg).value

    def get_hfss_prop(self):
        """
        Get HFSS properties.

        Returns
        -------
        material : str
            Name of the material property.
        solve_inside : bool
            Whether to solve inside.
        """
        prop_msg = self.__stub.GetHfssProp(self.msg)
        return prop_msg.material_name, prop_msg.solve_inside

    def remove_hfss_prop(self):
        """Remove HFSS properties."""
        self.__stub.RemoveHfssProp(self.msg)

    @property
    def is_zone_primitive(self):
        """:obj:`bool`: Flag indicating if the primitive object is a zone.

        This property is read-only.
        """
        return self.__stub.IsZonePrimitive(self.msg).value

    @property
    def can_be_zone_primitive(self):
        """:obj:`bool`: Flag indicating if the primitive can be a zone.

        This property is read-only.
        """
        return True

    def make_zone_primitive(self, zone_id):
        """Make the primitive a zone primitive with a zone specified by the provided ID.

        Parameters
        ----------
        zone_id : int
            ID of the zone primitive to use.
        """
        self.__stub.MakeZonePrimitive(messages.int_property_message(self, zone_id))


class Rectangle(Primitive):
    """Represents a rectangle object."""

    __stub: rectangle_pb2_grpc.RectangleServiceStub = StubAccessor(StubType.rectangle)

    @classmethod
    def create(
        cls, layout, layer, net, rep_type, param1, param2, param3, param4, corner_rad, rotation
    ):
        """Create a rectangle.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the rectangle in.
        layer : str or :class:`.Layer`
            Layer the rectangle is to created on.
        net : str or :class:`.Net` or None
            Net the rectangle is to have.
        rep_type : :class:`RectangleRepresentationType`
            Type that defines the meaning of the given parameters.
        param1 : :class:`.Value`
            X value of the lower-left point or center point.
        param2 : :class:`.Value`
            Y value of the lower-left point or center point.
        param3 : :class:`.Value`
            X value of the upper-right point or width.
        param4 : :class:`.Value`
            Y value of the upper-right point or height.
        corner_rad : :class:`.Value`
            Corner radius.
        rotation : :class:`.Value`
            Rotation.

        Returns
        -------
        Rectangle
            Rectangle created.
        """
        return Rectangle(
            cls.__stub.Create(
                rectangle_pb2.RectangleCreationMessage(
                    layout=layout.msg,
                    layer=messages.layer_ref_message(layer),
                    net=messages.net_ref_message(net),
                    representation_type=rep_type.value,
                    parameter1=messages.value_message(param1),
                    parameter2=messages.value_message(param2),
                    parameter3=messages.value_message(param3),
                    parameter4=messages.value_message(param4),
                    corner_radius=messages.value_message(corner_rad),
                    rotation=messages.value_message(rotation),
                )
            )
        )

    def get_parameters(self):
        """Get coordinate parameters.

        Returns
        -------
        tuple[
            :class:`RectangleRepresentationType`,
            :class:`.Value`,
            :class:`.Value`,
            :class:`.Value`,
            :class:`.Value`,
            :class:`.Value`,
            :class:`.Value`
        ]

            Returns a tuple in this format:

            **(representation_type, parameter1, parameter2, parameter3, parameter4, corner_radius, rotation)**

            **representation_type** : Type that defines given parameters meaning.

            **parameter1** : X value of lower left point or center point.

            **parameter2** : Y value of lower left point or center point.

            **parameter3** : X value of upper right point or width.

            **parameter4** : Y value of upper right point or height.

            **corner_radius** : Corner radius.

            **rotation** : Rotation.
        """
        rect_param_msg = self.__stub.GetParameters(self.msg)
        return (
            RectangleRepresentationType(rect_param_msg.representation_type),
            Value(rect_param_msg.parameter1),
            Value(rect_param_msg.parameter2),
            Value(rect_param_msg.parameter3),
            Value(rect_param_msg.parameter4),
            Value(rect_param_msg.corner_radius),
            Value(rect_param_msg.rotation),
        )

    def set_parameters(self, rep_type, param1, param2, param3, param4, corner_rad, rotation):
        """Set coordinate parameters.

        Parameters
        ----------
        rep_type : :class:`RectangleRepresentationType`
            Type that defines the meaning of the given parameters.
        param1 : :class:`.Value`
            X value of the lower-left point or center point.
        param2 : :class:`.Value`
            Y value of the lower-left point or center point.
        param3 : :class:`.Value`
            X value of the upper-right point or width.
        param4 : :class:`.Value`
            Y value of the upper-right point or height.
        corner_rad : :class:`.Value`
            Corner radius.
        rotation : :class:`.Value`
            Rotation.
        """
        self.__stub.SetParameters(
            rectangle_pb2.SetRectangleParametersMessage(
                target=self.msg,
                parameters=rectangle_pb2.RectangleParametersMessage(
                    representation_type=rep_type.value,
                    parameter1=messages.value_message(param1),
                    parameter2=messages.value_message(param2),
                    parameter3=messages.value_message(param3),
                    parameter4=messages.value_message(param4),
                    corner_radius=messages.value_message(corner_rad),
                    rotation=messages.value_message(rotation),
                ),
            )
        )

    @property
    def can_be_zone_primitive(self):
        """:obj:`bool`: Flag indicating if the rectangle can be a zone.

        This property is read-only.
        """
        return True

    @property
    def polygon_data(self):
        """:class:`.PolygonData`: \
        Polygon data object of the rectangle.

        This property is read-only.
        """
        return Rectangle.render(*self.get_parameters())

    @classmethod
    @parser.to_polygon_data
    def render(
        cls,
        rep_type,
        x_lower_left_or_center_x,
        y_lower_left_or_center_y,
        x_upper_right_or_width,
        y_upper_right_or_height,
        corner_radius,
        rotation,
        is_hole=False,
    ):
        """Get the polygon data of a rectangle.

        Parameters
        ----------
        rep_type : :class:`RectangleRepresentationType`
            Type that defines the meaning of the given parameters.
        x_lower_left_or_center_x : :class:`.Value`
            X value of the lower-left point or center point.
        y_lower_left_or_center_y : :class:`.Value`
            Y value of the lower-left point or center point.
        x_upper_right_or_width : :class:`.Value`
            X value of the upper-right point or width.
        y_upper_right_or_height : :class:`.Value`
            Y value of the upper-right point or height.
        corner_radius : :class:`.Value`
            Corner radius.
        rotation : :class:`.Value`
            Rotation.
        is_hole : bool, default: False
            Whether the rectangle is hole.

        Returns
        -------
        :class:`.PolygonData`
            Polygon data object created.
        """
        if rep_type == RectangleRepresentationType.LOWER_LEFT_UPPER_RIGHT:
            width = x_upper_right_or_width - x_lower_left_or_center_x
            height = y_upper_right_or_height - y_lower_left_or_center_y
            center_x = x_lower_left_or_center_x + width / 2.0
            center_y = y_lower_left_or_center_y + height / 2.0
            polygon_data = cls.__stub.Render(
                rectangle_pb2.RectanglePolygonDataMessage(
                    center_x=messages.value_message(center_x),
                    center_y=messages.value_message(center_y),
                    width=messages.value_message(width),
                    height=messages.value_message(height),
                    corner_radius=messages.value_message(corner_radius),
                    rotation=messages.value_message(rotation),
                    ishole=is_hole,
                )
            )
        elif rep_type == RectangleRepresentationType.CENTER_WIDTH_HEIGHT:
            polygon_data = cls.__stub.Render(
                rectangle_pb2.RectanglePolygonDataMessage(
                    center_x=messages.value_message(x_lower_left_or_center_x),
                    center_y=messages.value_message(y_lower_left_or_center_y),
                    width=messages.value_message(x_upper_right_or_width),
                    height=messages.value_message(y_upper_right_or_height),
                    corner_radius=messages.value_message(corner_radius),
                    rotation=messages.value_message(rotation),
                    ishole=is_hole,
                )
            )
        else:
            polygon_data = None
        return polygon_data


class Circle(Primitive):
    """Represents a circle object."""

    __stub: circle_pb2_grpc.CircleServiceStub = StubAccessor(StubType.circle)

    @classmethod
    def create(cls, layout, layer, net, center_x, center_y, radius):
        """Create a circle.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create this circle in.
        layer : str or :class:`.Layer`
            Layer to place the circle on.
        net : str or :class:`.Net` or None
            Net of the circle.
        center_x : :class:`.Value`
            X value of the center point.
        center_y : :class:`.Value`
            Y value of the center point.
        radius : :class:`.Value`
            Radius value of the circle.

        Returns
        -------
        Circle
            Circle created.
        """
        return Circle(
            cls.__stub.Create(
                circle_pb2.CircleCreationMessage(
                    layout=layout.msg,
                    layer=messages.layer_ref_message(layer),
                    net=messages.net_ref_message(net),
                    center_x=messages.value_message(center_x),
                    center_y=messages.value_message(center_y),
                    radius=messages.value_message(radius),
                )
            )
        )

    @classmethod
    @parser.to_polygon_data
    def render(cls, center_x, center_y, radius, is_hole):
        """Render a circle.

        Parameters
        ----------
        center_x : :class:`.Value`
            X value of the center point.
        center_y : :class:`.Value`
            Y value of the center point.
        radius : :class:`.Value`
            Radius value of the circle.
        is_hole: bool
            Whether the circle object is a hole.

        Returns
        -------
        :class:`.PolygonData`
            Circle created.
        """
        return cls.__stub.Render(
            circle_pb2.CircleRenderMessage(
                center_x=messages.value_message(center_x),
                center_y=messages.value_message(center_y),
                radius=messages.value_message(radius),
                is_hole=is_hole,
            )
        )

    def get_parameters(self):
        """Get parameters of the circle.

        Returns
        -------
        tuple[
            :class:`.Value`,
            :class:`.Value`,
            :class:`.Value`
        ]

            Returns a tuple in this format:

            **(center_x, center_y, radius)**

            **center_x** : X value of center point.

            **center_y** : Y value of center point.

            **radius** : Radius value of the circle.
        """
        circle_param_msg = self.__stub.GetParameters(self.msg)
        return (
            Value(circle_param_msg.center_x),
            Value(circle_param_msg.center_y),
            Value(circle_param_msg.radius),
        )

    def set_parameters(self, center_x, center_y, radius):
        """Set parameters of the circle.

         Parameters
         ----------
        center_x : :class:`.Value`
            X value of the center point.
        center_y : :class:`.Value`
            Y value of the center point.
        radius : :class:`.Value`
            Radius value of the circle.
        """
        self.__stub.SetParameters(
            circle_pb2.SetCircleParametersMessage(
                target=self.msg,
                parameters=circle_pb2.CircleParametersMessage(
                    center_x=messages.value_message(center_x),
                    center_y=messages.value_message(center_y),
                    radius=messages.value_message(radius),
                ),
            )
        )

    def get_polygon_data(self):
        """:class:`.PolygonData`: \
        Polygon data object of the circle."""
        return Circle.render(*self.get_parameters(), self.is_void)

    def can_be_zone_primitive(self):
        """:obj:`bool`: Flag indicating if a circle can be a zone."""
        return True


class Text(Primitive):
    """Represents a text object."""

    __stub: text_pb2_grpc.TextServiceStub = StubAccessor(StubType.text)

    @classmethod
    def create(cls, layout, layer, center_x, center_y, text):
        """Create a text object.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the text object in.
        layer : str or Layer
            Layer to place the text object on.
        center_x : :class:`.Value`
            X value of the center point.
        center_y : :class:`.Value`
            Y value of the center point.
        text: str
            Text string.

        Returns
        -------
        Text
            Text object created.
        """
        return Text(
            cls.__stub.Create(
                text_pb2.TextCreationMessage(
                    layout=layout.msg,
                    layer=messages.layer_ref_message(layer),
                    center_x=messages.value_message(center_x),
                    center_y=messages.value_message(center_y),
                    text=text,
                )
            )
        )

    def get_text_data(self):
        """Get the data for the text object.

        Returns
        -------
        tuple[
            :class:`.Value`,
            :class:`.Value`,
            str
        ]
            Returns a tuple in this format:

            **(center_x, center_y, text)**

            **center_x** : X value of center point.

            **center_y** : Y value of center point.

            **radius** : Text object's String value.
        """
        text_data_msg = self.__stub.GetTextData(self.msg)
        return (
            Value(text_data_msg.center_x),
            Value(text_data_msg.center_y),
            text_data_msg.text,
        )

    def set_text_data(self, center_x, center_y, text):
        """Set the data for the text object.

        Parameters
        ----------
        center_x : :class:`.Value`
            X value of the center point.
        center_y : :class:`.Value`
            Y value of the center point.
        text : str
            String value for the text object.
        """
        self.__stub.SetTextData(
            text_pb2.SetTextDataMessage(
                target=self.msg,
                data=text_pb2.TextDataMessage(
                    center_x=messages.value_message(center_x),
                    center_y=messages.value_message(center_y),
                    text=text,
                ),
            )
        )


class Polygon(Primitive):
    """Represents a polygon object."""

    __stub: polygon_pb2_grpc.PolygonServiceStub = StubAccessor(StubType.polygon)

    @classmethod
    def create(cls, layout, layer, net, polygon_data):
        """Create a polygon.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the polygon in.
        layer : str or :class:`.Layer`
            Layer to place the polygon on.
        net : str or :class:`.Net` or None
            Net of the polygon.
        polygon_data : :class:`.PolygonData`
            Outer contour of the polygon.

        Returns
        -------
        Polygon
            Polygon created.
        """
        return Polygon(
            cls.__stub.Create(
                polygon_pb2.PolygonCreationMessage(
                    layout=layout.msg,
                    layer=messages.layer_ref_message(layer),
                    net=messages.net_ref_message(net),
                    points=messages.polygon_data_message(polygon_data),
                )
            )
        )

    @property
    @parser.to_polygon_data
    def polygon_data(self):
        """:class:`.PolygonData`: Outer contour of the polygon."""
        return self.__stub.GetPolygonData(self.msg)

    @polygon_data.setter
    def polygon_data(self, poly):
        self.__stub.SetPolygonData(
            polygon_pb2.SetPolygonDataMessage(
                target=self.msg, poly=messages.polygon_data_message(poly)
            )
        )

    @property
    def can_be_zone_primitive(self):
        """:obj:`bool`: Flag indicating if a polygon can be a zone.

        This property is read-only.
        """
        return True


class Path(Primitive):
    """Represents a path object."""

    __stub: path_pb2_grpc.PathServiceStub = StubAccessor(StubType.path)

    @classmethod
    def create(cls, layout, layer, net, width, end_cap1, end_cap2, corner_style, points):
        """Create a path.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the path in.
        layer : str or :class:`.Layer`
            Layer to place the path on.
        net : str or :class:`.Net` or None
            Net of the path.
        width : :class:`.Value`
            Path width.
        end_cap1: :class:`.PathEndCapType`
            End cap style for the start of the path.
        end_cap2: :class:`.PathEndCapType`
            End cap style for the end of the path.
        corner_style : :class:`.PathCornerType`
            Corner style.
        points : :class:`.PolygonData`
            Centerline polygon data to set.

        Returns
        -------
        Path
            Path created.
        """
        return Path(
            cls.__stub.Create(
                path_pb2.PathCreationMessage(
                    layout=layout.msg,
                    layer=messages.layer_ref_message(layer),
                    net=messages.net_ref_message(net),
                    width=messages.value_message(width),
                    end_cap1=end_cap1.value,
                    end_cap2=end_cap2.value,
                    corner=corner_style.value,
                    points=messages.polygon_data_message(points),
                )
            )
        )

    @classmethod
    @parser.to_polygon_data
    def render(cls, width, end_cap1, end_cap2, corner_style, path):
        """Render a path.

        Parameters
        ----------
        width : :class:`.Value`
            Path width.
        end_cap1 : :class:`.PathEndCapType`
            End cap style for the start of the path.
        end_cap2 : :class:`.PathEndCapType`
            End cap style for the end of the path.
        corner_style : :class:`PathCornerType`
            Corner style.
        path : :class:`.PolygonData`
            Polygon data to set.

        Returns
        -------
        :class:`.PolygonData`
            Path rendered.
        """
        return cls.__stub.Render(
            path_pb2.PathRenderMessage(
                width=messages.value_message(width),
                end_cap1=end_cap1.value,
                end_cap2=end_cap2.value,
                corner_style=corner_style.value,
                path=messages.polygon_data_message(path),
            )
        )

    @property
    @parser.to_polygon_data
    def polygon_data(self):
        """:class:`.PolygonData`: Polygon data of this Path."""
        return self.__stub.GetPolygonData(self.msg)

    @property
    @parser.to_polygon_data
    def center_line(self):
        """:class:`.PolygonData`: Center line for the path."""
        return self.__stub.GetCenterLine(self.msg)

    @center_line.setter
    def center_line(self, center_line):
        path_pb2.SetCenterLineMessage(
            target=self.msg, center_line=messages.polygon_data_message(center_line)
        )

    def get_end_cap_style(self):
        """Get end cap styles for the path.

        Returns
        -------
        tuple[
            :class:`.PathEndCapType`,
            :class:`.PathEndCapType`
        ]

            Returns a tuple in this format:

            **(end_cap1, end_cap2)**

            **end_cap1** : End cap style of path start end cap.

            **end_cap2** : End cap style of path end end cap.
        """
        end_cap_msg = self.__stub.GetEndCapStyle(self.msg)
        return PathEndCapType(end_cap_msg.end_cap1), PathEndCapType(end_cap_msg.end_cap2)

    def set_end_cap_style(self, end_cap1, end_cap2):
        """Set end cap styles for the path.

        Parameters
        ----------
        end_cap1: :class:`.PathEndCapType`
            End cap style for the start of the path.
        end_cap2: :class:`.PathEndCapType`
            End cap style for the end of the path.
        """
        self.__stub.SetEndCapStyle(
            path_pb2.SetEndCapStyleMessage(
                target=self.msg,
                end_cap=path_pb2.EndCapStyleMessage(
                    end_cap1=end_cap1.value, end_cap2=end_cap2.value
                ),
            )
        )

    def get_clip_info(self):
        """Get the data used to clip the path.

        Returns
        -------
        tuple[:class:`.PolygonData`, bool]

            Returns a tuple in this format:

            **(clipping_poly, keep_inside)**

            **clipping_poly** : PolygonData used to clip the path.

            **keep_inside** : Indicates whether the part of the path inside the polygon is preserved.
        """
        clip_info_msg = self.__stub.GetClipInfo(self.msg)
        return (
            parser.to_polygon_data(clip_info_msg.clipping_poly),
            clip_info_msg.keep_inside,
        )

    def set_clip_info(self, clipping_poly, keep_inside=True):
        """Set the data used to clip the path.

        Parameters
        ----------
        clipping_poly : :class:`.PolygonData`
            Polygon data to use to clip the path.
        keep_inside: bool, default: True
            Whether the part of the path inside the polygon should be preserved.
        """
        self.__stub.SetClipInfo(
            path_pb2.SetClipInfoMessage(
                target=self.msg,
                clipping_poly=messages.polygon_data_message(clipping_poly),
                keep_inside=keep_inside,
            )
        )

    @property
    def corner_style(self):
        """:class:`PathCornerType`: Corner style of the path."""
        return PathCornerType(self.__stub.GetCornerStyle(self.msg).corner_style)

    @corner_style.setter
    def corner_style(self, corner_type):
        self.__stub.SetCornerStyle(
            path_pb2.SetCornerStyleMessage(
                target=self.msg,
                corner_style=path_pb2.CornerStyleMessage(corner_style=corner_type.value),
            )
        )

    @property
    def width(self):
        """:class:`.Value`: Path width."""
        return Value(self.__stub.GetWidth(self.msg).width)

    @width.setter
    def width(self, width):
        self.__stub.SetWidth(
            path_pb2.SetWidthMessage(
                target=self.msg,
                width=path_pb2.WidthMessage(width=messages.value_message(width)),
            )
        )

    @property
    def miter_ratio(self):
        """:class:`.Value`: Miter ratio."""
        return Value(self.__stub.GetMiterRatio(self.msg).miter_ratio)

    @miter_ratio.setter
    def miter_ratio(self, miter_ratio):
        self.__stub.SetMiterRatio(
            path_pb2.SetMiterRatioMessage(
                target=self.msg,
                miter_ratio=path_pb2.MiterRatioMessage(
                    miter_ratio=messages.value_message(miter_ratio)
                ),
            )
        )

    @property
    def can_be_zone_primitive(self):
        """:obj:`bool`: Flag indicating if the path can be a zone.

        This property is read-only.
        """
        return True


class Bondwire(Primitive):
    """Represents a bondwire object."""

    __stub: bondwire_pb2_grpc.BondwireServiceStub = StubAccessor(StubType.bondwire)

    @classmethod
    def create(
        cls,
        layout,
        bondwire_type,
        definition_name,
        placement_layer,
        width,
        material,
        start_context,
        start_layer_name,
        start_x,
        start_y,
        end_context,
        end_layer_name,
        end_x,
        end_y,
        net,
    ):
        """Create a bondwire.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the bondwire in.
        bondwire_type : :class:`BondwireType`
            Type of the bondwire. Options are ``kAPDBondWire`` and ``kJDECBondWire``.
        definition_name : str
            Bondwire definition name.
        placement_layer : str
            Layer name to create the bondwire on.
        width : :class:`.Value`
            Bondwire width.
        material : str
            Bondwire material name.
        start_context : :class:`.CellInstance`
            Start context. ``None`` means top-level,.
        start_layer_name : str
            Name of the start layer.
        start_x : :class:`.Value`
            X value of the start point.
        start_y : :class:`.Value`
            Y value of the start point.
        end_context : :class:`.CellInstance`
            End context: End content. ``None`` means top-level.
        end_layer_name : str
            Name of the end layer.
        end_x : :class:`.Value`
            X value of the end point.
        end_y : :class:`.Value`
            Y value of the end point.
        net : str or :class:`.Net` or None
            Net of the bondwire.

        Returns
        -------
        Bondwire
            Bondwire object created.
        """
        return Bondwire(
            cls.__stub.Create(
                bondwire_pb2.BondwireCreateMessage(
                    layout=layout.msg,
                    net=messages.net_ref_message(net),
                    bondwire_type=bondwire_type.value,
                    definition_name=definition_name,
                    placement_layer=placement_layer,
                    width=messages.value_message(width),
                    material=material,
                    start_context=messages.edb_obj_message(start_context),
                    start_layer_name=start_layer_name,
                    start_x=messages.value_message(start_x),
                    start_y=messages.value_message(start_y),
                    end_context=messages.edb_obj_message(end_context),
                    end_layer_name=end_layer_name,
                    end_x=messages.value_message(end_x),
                    end_y=messages.value_message(end_y),
                )
            )
        )

    def get_material(self, evaluated=True):
        """Get the material of the bondwire.

        Parameters
        ----------
        evaluated : bool, default: True
            Whether an evaluated (in variable namespace) material name is wanted.

        Returns
        -------
        str
            Material name.
        """
        return self.__stub.GetMaterial(Bondwire._bondwire_bool_message(self, evaluated))

    def set_material(self, material):
        """Set the material of the bondwire.

        Parameters
        ----------
        material : str
            Material name.
        """
        self.__stub.SetMaterial(bondwire_pb2.SetMaterialMessage(target=self.msg, material=material))

    @property
    def type(self):
        """:class:`BondwireType`: Type of the bondwire."""
        btype_msg = self.__stub.GetType(self.msg)
        return BondwireType(btype_msg.type)

    @type.setter
    def type(self, bondwire_type):
        self.__stub.SetType(
            bondwire_pb2.SetBondwireTypeMessage(target=self.msg, type=bondwire_type.value)
        )

    @property
    def cross_section_type(self):
        """:class:`BondwireCrossSectionType`:Cross-section type of the bondwire."""
        return BondwireCrossSectionType(self.__stub.GetCrossSectionType(self.msg).type)

    @cross_section_type.setter
    def cross_section_type(self, bondwire_type):
        self.__stub.SetCrossSectionType(
            bondwire_pb2.SetCrossSectionTypeMessage(target=self.msg, type=bondwire_type.value)
        )

    @property
    def cross_section_height(self):
        """:class:`.Value`: Cross-section height of the bondwire."""
        return Value(self.__stub.GetCrossSectionHeight(self.msg))

    @cross_section_height.setter
    def cross_section_height(self, height):
        self.__stub.SetCrossSectionHeight(
            bondwire_pb2.SetCrossSectionHeightMessage(
                target=self.msg, height=messages.value_message(height)
            )
        )

    def get_definition_name(self, evaluated=True):
        """Get the definition name of the bondwire object.

        Parameters
        ----------
        evaluated : bool, default: True
            Whether an evaluated (in variable namespace) material name is wanted.

        Returns
        -------
        str
            Bondwire definition name.
        """
        return self.__stub.GetDefinitionName(Bondwire._bondwire_bool_message(self, evaluated)).value

    def set_definition_name(self, definition_name):
        """Set the definition name of a bondwire.

        Parameters
        ----------
        definition_name : str
            Bondwire definition name to set.
        """
        self.__stub.SetDefinitionName(
            bondwire_pb2.SetDefinitionNameMessage(target=self.msg, definition_name=definition_name)
        )

    def get_traj(self):
        """Get trajectory parameters of the bondwire.

        Returns
        -------
        tuple[
            :class:`.Value`,
            :class:`.Value`,
            :class:`.Value`,
            :class:`.Value`
        ]

            Returns a tuple in this format:

            **(x1, y1, x2, y2)**

            **x1** : X value of the start point.

            **y1** : Y value of the start point.

            **x1** : X value of the end point.

            **y1** : Y value of the end point.
        """
        traj_msg = self.__stub.GetTraj(self.msg)
        return (
            Value(traj_msg.x1),
            Value(traj_msg.y1),
            Value(traj_msg.x2),
            Value(traj_msg.y2),
        )

    def set_traj(self, x1, y1, x2, y2):
        """Set the parameters of the trajectory of the bondwire.

        Parameters
        ----------
        x1 : :class:`.Value`
            X value of the the start point.
        y1 : :class:`.Value`
            Y value of the the start point.
        x2 : :class:`.Value`
            X value of the the end point.
        y2 : :class:`.Value`
            Y value of the end point.
        """
        self.__stub.SetTraj(
            target=self.msg,
            traj=bondwire_pb2.BondwireTrajMessage(
                x1=messages.value_message(x1),
                y1=messages.value_message(y1),
                x2=messages.value_message(x2),
                y2=messages.value_message(y2),
            ),
        )

    @property
    def width(self):
        """:class:`.Value`: Width of the bondwire."""
        val = self.__stub.GetWidthValue(self.msg)
        return Value(val)

    @width.setter
    def width(self, width):
        self.__stub.SetWidthValue(
            bondwire_pb2.BondwireValueMessage(target=self.msg, value=messages.value_message(width))
        )

    def get_start_elevation(self, start_context):
        """Get the start elevation layer of the bondwire.

        Parameters
        ----------
        start_context : :class:`.CellInstance`
            Start cell context of the bondwire.

        Returns
        -------
        :class:`.Layer`
            Start elevation level of the bondwire.
        """
        return Layer(
            self.__stub.GetStartElevation(Bondwire._get_elevation_message(self, start_context))
        ).cast()

    def set_start_elevation(self, start_context, layer):
        """Set the start elevation of the bondwire.

        Parameters
        ----------
        start_context : :class:`.CellInstance`
            Start cell context of the bondwire. ``None`` means top-level.
        layer : str or :class:`.Layer`
            Start layer of the bondwire.
        """
        self.__stub.SetStartElevation(Bondwire._set_elevation_message(self, start_context, layer))

    def get_end_elevation(self, end_context):
        """Get the end elevation layer of the bondwire.

        Parameters
        ----------
        end_context : :class:`.CellInstance`
            End cell context of the bondwire.

        Returns
        -------
        :class:`.Layer`
            End elevation layer of the bondwire.
        """
        return Layer(
            self.__stub.GetEndElevation(Bondwire._get_elevation_message(self, end_context))
        ).cast()

    def set_end_elevation(self, end_context, layer):
        """Set the end elevation of the bondwire.

        Parameters
        ----------
        end_context : :class:`.CellInstance`
            End cell context of the bondwire. ``None`` means top-level.
        layer : str or :class:`.Layer`
            End layer of the bondwire.
        """
        self.__stub.SetEndElevation(Bondwire._set_elevation_message(self, end_context, layer))

    @staticmethod
    def _bondwire_bool_message(b, evaluated):
        return bondwire_pb2.BondwireBoolMessage(target=b.msg, evaluated=evaluated)

    @staticmethod
    def _get_elevation_message(b, cell_instance):
        return bondwire_pb2.GetElevationMessage(
            bw=b.msg, cell_instance=messages.edb_obj_message(cell_instance)
        )

    @staticmethod
    def _set_elevation_message(b, cell_instance, lyrname):
        return bondwire_pb2.SetElevationMessage(
            target=Bondwire._get_elevation_message(b, cell_instance), lyrname=lyrname
        )


class PadstackInstance(Primitive):
    """Representis a padstack instance object."""

    __stub: padstack_instance_pb2_grpc.PadstackInstanceServiceStub = StubAccessor(
        StubType.padstack_instance
    )
    layout_obj_type = LayoutObjType.PADSTACK_INSTANCE

    @classmethod
    def create(
        cls,
        layout,
        net,
        name,
        padstack_def,
        position_x,
        position_y,
        rotation,
        top_layer,
        bottom_layer,
        solder_ball_layer,
        layer_map,
    ):
        """Create a padstack instance.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the padstack instance in.
        net : :class:`.Net`
            Net of the padstack instance.
        name : str
            Name of the padstack instance.
        padstack_def : :class:`.PadstackDef`
            Padstack definition of the padstack instance.
        position_x : :class:`.Value`
            Position x of the padstack instance.
        position_y : :class:`.Value`
            Position y of the padstack instance.
        rotation : :class:`.Value`
            Rotation of the padstack instance.
        top_layer : :class:`.Layer`
            Top layer of the padstack instance.
        bottom_layer : :class:`.Layer`
            Bottom layer of the padstack instance.
        solder_ball_layer : :class:`.Layer`
            Solder ball layer of the padstack instance or ``None`` for none.
        layer_map : :class:`.LayerMap`
            Layer map of the padstack instance. ``None`` or empty results in
            auto-mapping.

        Returns
        -------
        PadstackInstance
            Padstack instance created.
        """
        padstack_instance = PadstackInstance(
            cls.__stub.Create(
                padstack_instance_pb2.PadstackInstCreateMessage(
                    layout=layout.msg,
                    net=net.msg,
                    name=name,
                    padstack_def=padstack_def.msg,
                    rotation=messages.value_message(rotation),
                    top_layer=top_layer.msg,
                    bottom_layer=bottom_layer.msg,
                    solder_ball_layer=messages.edb_obj_message(solder_ball_layer),
                    layer_map=messages.edb_obj_message(layer_map),
                )
            )
        )
        padstack_instance.set_position_and_rotation(position_x, position_y, rotation)
        return padstack_instance

    @property
    def padstack_def(self):
        """:class:`.PadstackDef`: \
        Definition of the padstack instance."""
        return PadstackDef(self.__stub.GetPadstackDef(self.msg))

    @property
    def name(self):
        """:obj:`str`: Name of the padstack instance."""
        return self.__stub.GetName(self.msg).value

    @name.setter
    def name(self, name):
        self.__stub.SetName(messages.edb_obj_name_message(self, name))

    def get_position_and_rotation(self):
        """Get the position and rotation of the padstack instance.

        Returns
        -------
        tuple[
            :class:`.Value`,
            :class:`.Value`,
            :class:`.Value`
        ]

            Returns a tuple in this format:

            **(x, y, rotation)**

            **x** : X coordinate.

            **y** : Y coordinate.

            **rotation** : Rotation in radians.
        """
        params = self.__stub.GetPositionAndRotation(self.msg)
        return (
            Value(params.position.x),
            Value(params.position.y),
            Value(params.rotation),
        )

    def set_position_and_rotation(self, x, y, rotation):
        """Set the position and rotation of the padstack instance.

        Parameters
        ----------
        x : :class:`.Value`
            x : X coordinate.
        y : :class:`.Value`
            y : Y coordinate.
        rotation : :class:`.Value`
            Rotation in radians.
        """
        self.__stub.SetPositionAndRotation(
            padstack_instance_pb2.PadstackInstSetPositionAndRotationMessage(
                target=self.msg,
                params=padstack_instance_pb2.PadstackInstPositionAndRotationMessage(
                    position=messages.point_message((x, y)),
                    rotation=messages.value_message(rotation),
                ),
            )
        )

    def get_layer_range(self):
        """Get the top and bottom layers of the padstack instance.

        Returns
        -------
        tuple[:class:`.Layer`, :class:`.Layer`]
            The tuple is in this format: ``(top_layer, bottom_layer)``.

            - ``top_layer``: Top layer of the padstack instance
            - ``bottom_layer``: Bottom layer of the padstack instance
        """
        params = self.__stub.GetLayerRange(self.msg)
        return (
            Layer(params.top_layer).cast(),
            Layer(params.bottom_layer).cast(),
        )

    def set_layer_range(self, top_layer, bottom_layer):
        """Set the top and bottom layers of the padstack instance.

        Parameters
        ----------
        top_layer : :class:`.Layer`
            Top layer of the padstack instance.
        bottom_layer : :class:`.Layer`
            Bottom layer of the padstack instance.
        """
        self.__stub.SetLayerRange(
            padstack_instance_pb2.PadstackInstSetLayerRangeMessage(
                target=self.msg,
                range=padstack_instance_pb2.PadstackInstLayerRangeMessage(
                    top_layer=top_layer.msg,
                    bottom_layer=bottom_layer.msg,
                ),
            )
        )

    @property
    def solderball_layer(self):
        """:class:`.Layer`: Solderball layer of the padstack instance."""
        return Layer(self.__stub.GetSolderBallLayer(self.msg)).cast()

    @solderball_layer.setter
    def solderball_layer(self, solderball_layer):
        self.__stub.SetSolderBallLayer(
            padstack_instance_pb2.PadstackInstSetSolderBallLayerMessage(
                target=self.msg,
                layer=solderball_layer.msg,
            )(self, solderball_layer)
        )

    @property
    def layer_map(self):
        """:class:`.LayerMap`: Layer map of the padstack instance."""
        return LayerMap(self.__stub.GetLayerMap(self.msg))

    @layer_map.setter
    def layer_map(self, layer_map):
        self.__stub.SetLayerMap(messages.pointer_property_message(self, layer_map))

    def get_hole_overrides(self):
        """Get the hole overrides of the padstack instance.

        Returns
        -------
        tuple[
            bool,
            :class:`.Value`
        ]

            Returns a tuple in this format:

            **(is_hole_override, hole_override)**

            **is_hole_override** : If padstack instance is hole override.

            **hole_override** : Hole override diameter of this padstack instance.
        """
        params = self.__stub.GetHoleOverrides(self.msg)
        return (
            params.is_hole_override,
            Value(params.hole_override),
        )

    def set_hole_overrides(self, is_hole_override, hole_override):
        """Set the hole overrides of the padstack instance.

        Parameters
        ----------
        is_hole_override : bool
            Whether the padstack instance is a hole override.
        hole_override : :class:`.Value`
            Hole override diameter of the padstack instance.
        """
        self.__stub.SetHoleOverrides(
            padstack_instance_pb2.PadstackInstSetHoleOverridesMessage(
                target=self.msg,
                hole_override_msg=padstack_instance_pb2.PadstackInstHoleOverridesMessage(
                    is_hole_override=is_hole_override,
                    hole_override=messages.value_message(hole_override),
                ),
            )
        )

    @property
    def is_layout_pin(self):
        """:obj:`bool`: Flag indicating if the padstack instance is a layout pin."""
        return self.__stub.GetIsLayoutPin(self.msg).value

    @is_layout_pin.setter
    def is_layout_pin(self, is_layout_pin):
        self.__stub.SetIsLayoutPin(
            padstack_instance_pb2.PadstackInstSetIsLayoutPinMessage(
                target=self.msg,
                is_layout_pin=is_layout_pin,
            )
        )

    def get_back_drill_type(self, from_bottom):
        """Get the back drill type of the padstack instance.

        Parameters
        ----------
        from_bottom : bool
            Whether to get the back drill type from the bottom.

        Returns
        -------
        :class:`BackDrillType`
            Back drill type of the padastack instance.
        """
        return BackDrillType(
            self.__stub.GetBackDrillType(
                PadstackInstance._get_back_drill_message(self, from_bottom)
            ).type
        )

    def get_back_drill_by_layer(self, from_bottom):
        """Get the back drill type by the layer.

        Parameters
        ----------
        from_bottom : bool
            Whether to get the back drill type from the bottom.

        Returns
        -------
        tuple[
            bool,
            :class:`.Value`,
            :class:`.Value`
        ]

            Returns a tuple in this format:

            **(drill_to_layer, offset, diameter)**

            **drill_to_layer** : Layer drills to. If drill from top, drill stops at the upper elevation of the layer.\
            If from bottom, drill stops at the lower elevation of the layer.

            **offset** : Layer offset (or depth if layer is empty).

            **diameter** : Drilling diameter.
        """
        params = self.__stub.GetBackDrillByLayer(
            PadstackInstance._get_back_drill_message(self, from_bottom)
        )

        return (
            Layer(params.drill_to_layer).cast(),
            Value(params.offset),
            Value(params.diameter),
        )

    def set_back_drill_by_layer(self, drill_to_layer, offset, diameter, from_bottom):
        """Set the back drill by the layer.

        Parameters
        ----------
        drill_to_layer : :class:`.Layer`
            Layer to drill to. If drilling from the top, the drill stops at the upper
            elevation of the layer. If drilling from the bottom, the drill stops at
            the lower elevation of the layer.
        offset : :class:`.Value`
            Layer offset (or depth if the layer is empty).
        diameter : :class:`.Value`
            Drilling diameter.
        from_bottom : bool
            Whether to set the back drill type from the bottom.
        """
        self.__stub.SetBackDrillByLayer(
            padstack_instance_pb2.PadstackInstSetBackDrillByLayerMessage(
                target=self.msg,
                drill_to_layer=drill_to_layer.msg,
                offset=messages.value_message(offset),
                diameter=messages.value_message(diameter),
                from_bottom=from_bottom,
            )
        )

    def get_back_drill_by_depth(self, from_bottom):
        """Get the back drill type by depth.

        Parameters
        ----------
        from_bottom : bool
            Whether to get the back drill type from the bottom.

        Returns
        -------
        tuple[
            bool,
            :class:`.Value`
        ]
            Returns a tuple in this format:

            **(drill_depth, diameter)**

            **drill_depth** : Drilling depth, may not align with layer.

            **diameter** : Drilling diameter.
        """
        params = self.__stub.GetBackDrillByDepth(
            PadstackInstance._get_back_drill_message(self, from_bottom)
        )
        return Value(params.drill_depth), Value(params.diameter)

    def set_back_drill_by_depth(self, drill_depth, diameter, from_bottom):
        """Set the back drill type by depth.

        Parameters
        ----------
        drill_depth : :class:`.Value`
            Drilling depth, which may not align with the layer.
        diameter : :class:`.Value`
            Drilling diameter.
        from_bottom : bool
            Whether to set the back drill type from the bottom.
        """
        self.__stub.SetBackDrillByDepth(
            padstack_instance_pb2.PadstackInstSetBackDrillByDepthMessage(
                target=self.msg,
                drill_depth=messages.value_message(drill_depth),
                diameter=messages.value_message(diameter),
                from_bottom=from_bottom,
            )
        )

    def get_padstack_instance_terminal(self):
        """:class:`.TerminalInstance`: \
        Terminal of the padstack instance."""
        from ansys.edb.core.terminal import terminals

        return terminals.TerminalInstance(self.__stub.GetPadstackInstanceTerminal(self.msg))

    def is_in_pin_group(self, pin_group):
        """Determine if the padstack instance is in a given pin group.

        Parameters
        ----------
        pin_group : :class:`.PinGroup`
            Pin group to check if the padstack instance is in it.

        Returns
        -------
        bool
            Whether the padstack instance is in a pin group.
        """
        return self.__stub.IsInPinGroup(
            padstack_instance_pb2.PadstackInstIsInPinGroupMessage(
                target=self.msg,
                pin_group=pin_group.msg,
            )
        ).value

    @property
    def pin_groups(self):
        """:obj:`list` of :class:`.PinGroup`: \
        Pin groups of the padstack instance.

        This property is read-only.
        """
        from ansys.edb.core.hierarchy import pin_group

        pins = self.__stub.GetPinGroups(self.msg).items
        return [pin_group.PinGroup(p) for p in pins]

    @staticmethod
    def _get_back_drill_message(padstack_inst, from_bottom):
        return padstack_instance_pb2.PadstackInstGetBackDrillMessage(
            target=padstack_inst.msg,
            from_bottom=from_bottom,
        )


class BoardBendDef(Primitive):
    """Represents a board bend definition instance."""

    __stub: board_bend_def_pb2_grpc.BoardBendDefServiceStub = StubAccessor(StubType.board_bend_def)

    @classmethod
    def create(cls, layout, zone_prim, bend_middle, bend_radius, bend_angle):
        """Create a board bend definition.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the board bend definition in.
        zone_prim : :class:`.Primitive`
            Zone primitive to create the board bend definition on.
        bend_middle : (:term:`Point2DLike`, :term:`Point2DLike`)
            Tuple containing the starting and ending points of the line that represents
            the middle of the bend.
        bend_radius : :term:`ValueLike`
            Radius of the bend.
        bend_angle : :term:`ValueLike`
            Angle of the bend.

        Returns
        -------
        BoardBendDef
            Board bend definition created.
        """
        return BoardBendDef(
            cls.__stub.Create(
                board_bend_def_pb2.BoardBendDefCreateMessage(
                    layout=layout.msg,
                    zone_prim=zone_prim.msg,
                    middle=messages.point_pair_message(bend_middle),
                    radius=messages.value_message(bend_radius),
                    angle=messages.value_message(bend_angle),
                )
            )
        )

    @property
    def boundary_primitive(self):
        """:class:`.Primitive`: Zone primitive the board bend is placed on.

        This property is read-only.
        """
        return Primitive(self.__stub.GetBoundaryPrim(self.msg)).cast()

    @property
    @parser.to_point_data_pair
    def bend_middle(self):
        """(:term:`Point2DLike`, :term:`Point2DLike`): Tuple of the bend middle based on starting and ending points."""
        return self.__stub.GetBendMiddle(self.msg)

    @bend_middle.setter
    def bend_middle(self, bend_middle):
        self.__stub.SetBendMiddle(messages.point_pair_property_message(self, bend_middle))

    @property
    def radius(self):
        """:term:`ValueLike`: Radius of the bend."""
        return Value(self.__stub.GetRadius(self.msg))

    @radius.setter
    def radius(self, val):
        self.__stub.SetRadius(messages.value_property_message(self, val))

    @property
    def angle(self):
        """:term:`ValueLike`: Angle of the bend."""
        return Value(self.__stub.GetAngle(self.msg))

    @angle.setter
    def angle(self, val):
        self.__stub.SetAngle(messages.value_property_message(self, val))

    @property
    @parser.to_polygon_data_list
    def bent_regions(self):
        """:obj:`list` of :class:`.PolygonData`: Bent region polygons.

        This list of a collection of polygon data represents the areas bent by the bend definition.

        This property is read-only.
        """
        return self.__stub.GetBentRegions(self.msg)
